import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union

import evaluate
import torch
from datasets import Dataset, DatasetDict
from packaging.version import parse as parse_version
from sentence_transformers import SentenceTransformerTrainer, losses
from sentence_transformers.losses.BatchHardTripletLoss import BatchHardTripletLossDistanceFunction
from sentence_transformers.model_card import ModelCardCallback as STModelCardCallback
from sentence_transformers.training_args import BatchSamplers, SentenceTransformerTrainingArguments
from sklearn.preprocessing import LabelEncoder
from torch import nn
from transformers import __version__ as transformers_version
from transformers.integrations import CodeCarbonCallback
from transformers.trainer_callback import IntervalStrategy, TrainerCallback
from transformers.trainer_utils import HPSearchBackend, default_compute_objective, number_of_arguments, set_seed
from transformers.utils.import_utils import is_in_notebook

from setfit.model_card import ModelCardCallback

from . import logging
from .integrations import default_hp_search_backend, is_optuna_available, run_hp_search_optuna
from .losses import SupConLoss
from .sampler import ContrastiveDataset
from .training_args import TrainingArguments
from .utils import BestRun, default_hp_space_optuna


if TYPE_CHECKING:
    import optuna

    from .modeling import SetFitModel

logging.set_verbosity_info()
logger = logging.get_logger(__name__)


class BCSentenceTransformersTrainer(SentenceTransformerTrainer):
    """
    Subclass of SentenceTransformerTrainer that is backwards compatible with the SetFit API.
    """

    def __init__(
        self, setfit_model: "SetFitModel", setfit_args: "TrainingArguments", callbacks: List[TrainerCallback], **kwargs
    ):
        self._setfit_model = setfit_model
        self._setfit_args = setfit_args
        self.logs_prefix = "embedding"
        super().__init__(
            model=setfit_model.model_body,
            args=SentenceTransformerTrainingArguments(output_dir=setfit_args.output_dir),
            **kwargs,
        )
        self._apply_training_arguments(setfit_args)

        for callback in list(self.callback_handler.callbacks):
            if isinstance(callback, CodeCarbonCallback):
                self.setfit_model.model_card_data.code_carbon_callback = callback

            if isinstance(callback, STModelCardCallback):
                self.remove_callback(callback)

        if is_in_notebook():
            from transformers.utils.notebook import NotebookProgressCallback

            from setfit.notebook import SetFitNotebookProgressCallback

            if self.pop_callback(NotebookProgressCallback):
                self.add_callback(SetFitNotebookProgressCallback)

        def overwritten_call_event(self, event, args, state, control, **kwargs):
            for callback in self.callbacks:
                result = getattr(callback, event)(
                    self.setfit_args,
                    state,
                    control,
                    model=self.setfit_model,
                    st_model=self.model,
                    st_args=args,
                    tokenizer=(
                        self.processing_class
                        if parse_version(transformers_version) >= parse_version("4.46.0")
                        else self.tokenizer
                    ),
                    optimizer=self.optimizer,
                    lr_scheduler=self.lr_scheduler,
                    train_dataloader=self.train_dataloader,
                    eval_dataloader=self.eval_dataloader,
                    **kwargs,
                )
                # A Callback can skip the return of `control` if it doesn't change it.
                if result is not None:
                    control = result
            return control

        self.callback_handler.call_event = lambda *args, **kwargs: overwritten_call_event(
            self.callback_handler, *args, **kwargs
        )
        self.callback_handler.setfit_args = setfit_args
        self.callback_handler.setfit_model = setfit_model
        for callback in callbacks:
            self.add_callback(callback)
        self.callback_handler.on_init_end(self.args, self.state, self.control)

    def add_model_card_callback(self, *args, **kwargs):
        pass

    @property
    def setfit_model(self) -> "SetFitModel":
        return self._setfit_model

    @setfit_model.setter
    def setfit_model(self, setfit_model: "SetFitModel") -> None:
        self._setfit_model = setfit_model
        self.model = setfit_model.model_body
        self.callback_handler.setfit_model = setfit_model

    @property
    def setfit_args(self) -> TrainingArguments:
        return self._setfit_args

    @setfit_args.setter
    def setfit_args(self, setfit_args: TrainingArguments) -> None:
        self._setfit_args = setfit_args
        self._apply_training_arguments(setfit_args)
        self.callback_handler.setfit_args = setfit_args

    def _apply_training_arguments(self, args: TrainingArguments) -> None:
        """
        Propagate the SetFit TrainingArguments to the SentenceTransformer Trainer.
        """
        self.args.output_dir = args.output_dir

        self.args.per_device_train_batch_size = args.embedding_batch_size
        self.args.per_device_eval_batch_size = args.embedding_batch_size
        self.args.num_train_epochs = args.embedding_num_epochs
        self.args.max_steps = args.max_steps
        self.args.learning_rate = args.body_embedding_learning_rate
        self.args.fp16 = args.use_amp
        self.args.weight_decay = args.l2_weight
        self.args.seed = args.seed
        self.args.report_to = args.report_to
        self.args.run_name = args.run_name
        self.args.warmup_ratio = args.warmup_proportion

        self.args.logging_dir = args.logging_dir
        self.args.logging_strategy = args.logging_strategy
        self.args.logging_first_step = args.logging_first_step
        self.args.logging_steps = args.logging_steps

        self.args.eval_strategy = args.eval_strategy
        self.args.eval_steps = args.eval_steps
        self.args.eval_delay = args.eval_delay

        self.args.save_strategy = args.save_strategy
        self.args.save_steps = args.save_steps
        self.args.save_total_limit = args.save_total_limit

        self.args.load_best_model_at_end = args.load_best_model_at_end
        self.args.metric_for_best_model = args.metric_for_best_model
        self.args.greater_is_better = args.greater_is_better

    def _set_logs_prefix(self, logs_prefix: str) -> None:
        """Set the logging prefix.

        Args:
            logs_mapper (str): The logging prefix, e.g. "aspect_embedding".
        """
        self.logs_prefix = logs_prefix

    def log(self, logs: Dict[str, float], *args, **kwargs) -> None:
        logs = {f"{self.logs_prefix}_{k}" if k == "loss" else k: v for k, v in logs.items()}
        return super().log(logs, *args, **kwargs)

    def evaluate(
        self,
        eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> Dict[str, float]:
        return super().evaluate(
            eval_dataset=eval_dataset,
            ignore_keys=ignore_keys,
            metric_key_prefix=f"{metric_key_prefix}_{self.logs_prefix}",
        )


class ColumnMappingMixin:
    _REQUIRED_COLUMNS = {"text", "label"}

    def _validate_column_mapping(self, dataset: "Dataset") -> None:
        """
        Validates the provided column mapping against the dataset.
        """
        column_names = set(dataset.column_names)
        if self.column_mapping is None and not self._REQUIRED_COLUMNS.issubset(column_names):
            # Issue #226: load_dataset will automatically assign points to "train" if no split is specified
            if column_names == {"train"} and isinstance(dataset, DatasetDict):
                raise ValueError(
                    "SetFit expected a Dataset, but it got a DatasetDict with the split ['train']. "
                    "Did you mean to select the training split with dataset['train']?"
                )
            elif isinstance(dataset, DatasetDict):
                raise ValueError(
                    f"SetFit expected a Dataset, but it got a DatasetDict with the splits {sorted(column_names)}. "
                    "Did you mean to select one of these splits from the dataset?"
                )
            else:
                raise ValueError(
                    f"SetFit expected the dataset to have the columns {sorted(self._REQUIRED_COLUMNS)}, "
                    f"but only the columns {sorted(column_names)} were found. "
                    "Either make sure these columns are present, or specify which columns to use with column_mapping in Trainer."
                )
        if self.column_mapping is not None:
            missing_columns = set(self._REQUIRED_COLUMNS)
            # Remove columns that will be provided via the column mapping
            missing_columns -= set(self.column_mapping.values())
            # Remove columns that will be provided because they are in the dataset & not mapped away
            missing_columns -= set(dataset.column_names) - set(self.column_mapping.keys())
            if missing_columns:
                raise ValueError(
                    f"The following columns are missing from the column mapping: {missing_columns}. "
                    "Please provide a mapping for all required columns."
                )
            if not set(self.column_mapping.keys()).issubset(column_names):
                raise ValueError(
                    f"The column mapping expected the columns {sorted(self.column_mapping.keys())} in the dataset, "
                    f"but the dataset had the columns {sorted(column_names)}."
                )

    def _apply_column_mapping(self, dataset: "Dataset", column_mapping: Dict[str, str]) -> "Dataset":
        """
        Applies the provided column mapping to the dataset, renaming columns accordingly.
        Extra features not in the column mapping are prefixed with `"feat_"`.
        """
        dataset = dataset.rename_columns(
            {
                **column_mapping,
                **{
                    col: f"feat_{col}"
                    for col in dataset.column_names
                    if col not in column_mapping and col not in self._REQUIRED_COLUMNS
                },
            }
        )
        dset_format = dataset.format
        dataset = dataset.with_format(
            type=dset_format["type"],
            columns=dataset.column_names,
            output_all_columns=dset_format["output_all_columns"],
            **dset_format["format_kwargs"],
        )
        return dataset


class Trainer(ColumnMappingMixin):
    """Trainer to train a SetFit model.

    Args:
        model (`SetFitModel`, *optional*):
            The model to train. If not provided, a `model_init` must be passed.
        args (`TrainingArguments`, *optional*):
            The training arguments to use.
        train_dataset (`Dataset`):
            The training dataset.
        eval_dataset (`Dataset`, *optional*):
            The evaluation dataset.
        model_init (`Callable[[], SetFitModel]`, *optional*):
            A function that instantiates the model to be used. If provided, each call to
            [`Trainer.train`] will start from a new instance of the model as given by this
            function when a `trial` is passed.
        metric (`str` or `Callable`, *optional*, defaults to `"accuracy"`):
            The metric to use for evaluation. If a string is provided, we treat it as the metric
            name and load it with default settings. If a callable is provided, it must take two arguments
            (`y_pred`, `y_test`) and return a dictionary with metric keys to values.
        metric_kwargs (`Dict[str, Any]`, *optional*):
            Keyword arguments passed to the evaluation function if `metric` is an evaluation string like "f1".
            For example useful for providing an averaging strategy for computing f1 in a multi-label setting.
        callbacks (`List[`[`~transformers.TrainerCallback`]`]`, *optional*):
            A list of callbacks to customize the training loop. Will add those to the list of default callbacks
            detailed in [here](https://huggingface.co/docs/transformers/main/en/main_classes/callback).
            If you want to remove one of the default callbacks used, use the [`Trainer.remove_callback`] method.
        column_mapping (`Dict[str, str]`, *optional*):
            A mapping from the column names in the dataset to the column names expected by the model.
            The expected format is a dictionary with the following format:
            `{"text_column_name": "text", "label_column_name": "label"}`.
    """

    def __init__(
        self,
        model: Optional["SetFitModel"] = None,
        args: Optional[TrainingArguments] = None,
        train_dataset: Optional["Dataset"] = None,
        eval_dataset: Optional["Dataset"] = None,
        model_init: Optional[Callable[[], "SetFitModel"]] = None,
        metric: Union[str, Callable[["Dataset", "Dataset"], Dict[str, float]]] = "accuracy",
        metric_kwargs: Optional[Dict[str, Any]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        column_mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        if args is not None and not isinstance(args, TrainingArguments):
            raise ValueError("`args` must be a `TrainingArguments` instance imported from `setfit`.")
        self.args = args or TrainingArguments()
        self.column_mapping = column_mapping
        if train_dataset:
            self._validate_column_mapping(train_dataset)
            if self.column_mapping is not None:
                logger.info("Applying column mapping to the training dataset")
                train_dataset = self._apply_column_mapping(train_dataset, self.column_mapping)
        self.train_dataset = train_dataset

        if eval_dataset:
            self._validate_column_mapping(eval_dataset)
            if self.column_mapping is not None:
                logger.info("Applying column mapping to the evaluation dataset")
                eval_dataset = self._apply_column_mapping(eval_dataset, self.column_mapping)
        self.eval_dataset = eval_dataset

        self.model_init = model_init
        self.metric = metric
        self.metric_kwargs = metric_kwargs
        self.logs_mapper = {}

        # Seed must be set before instantiating the model when using model_init.
        set_seed(args.seed if args is not None else 12)

        if model is None:
            if model_init is not None:
                model = self.call_model_init()
            else:
                raise RuntimeError("`Trainer` requires either a `model` or `model_init` argument.")
        else:
            if model_init is not None:
                raise RuntimeError("`Trainer` requires either a `model` or `model_init` argument, but not both.")

        self._model = model
        self.hp_search_backend = None

        callbacks = callbacks + [ModelCardCallback(self)] if callbacks else [ModelCardCallback(self)]
        self.st_trainer = BCSentenceTransformersTrainer(
            setfit_model=model,
            setfit_args=self.args,
            callbacks=callbacks,
        )

    @property
    def args(self) -> TrainingArguments:
        return self._args

    @args.setter
    def args(self, args: TrainingArguments) -> None:
        self._args = args
        if hasattr(self, "st_trainer"):
            self.st_trainer.setfit_args = args

    @property
    def model(self) -> "SetFitModel":
        return self._model

    @model.setter
    def model(self, model: "SetFitModel") -> None:
        self._model = model
        if hasattr(self, "st_trainer"):
            self.st_trainer.setfit_model = model

    def add_callback(self, callback: Union[type, TrainerCallback]) -> None:
        """
        Add a callback to the current list of [`~transformers.TrainerCallback`].

        Args:
           callback (`type` or [`~transformers.TrainerCallback`]):
               A [`~transformers.TrainerCallback`] class or an instance of a [`~transformers.TrainerCallback`]. In the
               first case, will instantiate a member of that class.
        """
        self.st_trainer.add_callback(callback)

    def pop_callback(self, callback: Union[type, TrainerCallback]) -> TrainerCallback:
        """
        Remove a callback from the current list of [`~transformers.TrainerCallback`] and returns it.

        If the callback is not found, returns `None` (and no error is raised).

        Args:
           callback (`type` or [`~transformers.TrainerCallback`]):
               A [`~transformers.TrainerCallback`] class or an instance of a [`~transformers.TrainerCallback`]. In the
               first case, will pop the first member of that class found in the list of callbacks.

        Returns:
            [`~transformers.TrainerCallback`]: The callback removed, if found.
        """
        return self.st_trainer.pop_callback(callback)

    def remove_callback(self, callback: Union[type, TrainerCallback]) -> None:
        """
        Remove a callback from the current list of [`~transformers.TrainerCallback`].

        Args:
           callback (`type` or [`~transformers.TrainerCallback`]):
               A [`~transformers.TrainerCallback`] class or an instance of a [`~transformers.TrainerCallback`]. In the
               first case, will remove the first member of that class found in the list of callbacks.
        """
        self.st_trainer.remove_callback(callback)

    def apply_hyperparameters(self, params: Dict[str, Any], final_model: bool = False) -> None:
        """Applies a dictionary of hyperparameters to both the trainer and the model

        Args:
            params (`Dict[str, Any]`): The parameters, usually from `BestRun.hyperparameters`
            final_model (`bool`, *optional*, defaults to `False`): If `True`, replace the `model_init()` function with a fixed model based on the parameters.
        """

        if self.args is not None:
            self.args = self.args.update(params, ignore_extra=True)
        else:
            self.args = TrainingArguments.from_dict(params, ignore_extra=True)

        # Seed must be set before instantiating the model when using model_init.
        set_seed(self.args.seed)
        self.model = self.model_init(params)
        if final_model:
            self.model_init = None

    def _hp_search_setup(self, trial: Union["optuna.Trial", Dict[str, Any]]) -> None:
        """HP search setup code"""

        # Heavily inspired by transformers.Trainer._hp_search_setup
        if self.hp_search_backend is None or trial is None:
            return

        if isinstance(trial, Dict):  # For passing a Dict to train() -- mostly unused for now
            params = trial
        elif self.hp_search_backend == HPSearchBackend.OPTUNA:
            params = self.hp_space(trial)
        else:
            raise ValueError("Invalid trial parameter")

        logger.info(f"Trial: {params}")
        self.apply_hyperparameters(params, final_model=False)

    def call_model_init(self, params: Optional[Dict[str, Any]] = None) -> "SetFitModel":
        model_init_argcount = number_of_arguments(self.model_init)
        if model_init_argcount == 0:
            model = self.model_init()
        elif model_init_argcount == 1:
            model = self.model_init(params)
        else:
            raise RuntimeError("`model_init` should have 0 or 1 argument.")

        if model is None:
            raise RuntimeError("`model_init` should not return None.")

        return model

    def freeze(self, component: Optional[Literal["body", "head"]] = None) -> None:
        """Freeze the model body and/or the head, preventing further training on that component until unfrozen.

        This method is deprecated, use `SetFitModel.freeze` instead.

        Args:
            component (`Literal["body", "head"]`, *optional*): Either "body" or "head" to freeze that component.
                If no component is provided, freeze both. Defaults to None.
        """
        warnings.warn(
            f"`{self.__class__.__name__}.freeze` is deprecated and will be removed in v2.0.0 of SetFit. "
            "Please use `SetFitModel.freeze` directly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.model.freeze(component)

    def unfreeze(
        self, component: Optional[Literal["body", "head"]] = None, keep_body_frozen: Optional[bool] = None
    ) -> None:
        """Unfreeze the model body and/or the head, allowing further training on that component.

        This method is deprecated, use `SetFitModel.unfreeze` instead.

        Args:
            component (`Literal["body", "head"]`, *optional*): Either "body" or "head" to unfreeze that component.
                If no component is provided, unfreeze both. Defaults to None.
            keep_body_frozen (`bool`, *optional*): Deprecated argument, use `component` instead.
        """
        warnings.warn(
            f"`{self.__class__.__name__}.unfreeze` is deprecated and will be removed in v2.0.0 of SetFit. "
            "Please use `SetFitModel.unfreeze` directly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.model.unfreeze(component, keep_body_frozen=keep_body_frozen)

    def train(
        self,
        args: Optional[TrainingArguments] = None,
        trial: Optional[Union["optuna.Trial", Dict[str, Any]]] = None,
        **kwargs,
    ) -> None:
        """
        Main training entry point.

        Args:
            args (`TrainingArguments`, *optional*):
                Temporarily change the training arguments for this training call.
            trial (`optuna.Trial` or `Dict[str, Any]`, *optional*):
                The trial run or the hyperparameter dictionary for hyperparameter search.
        """
        if len(kwargs):
            warnings.warn(
                f"`{self.__class__.__name__}.train` does not accept keyword arguments anymore. "
                f"Please provide training arguments via a `TrainingArguments` instance to the `{self.__class__.__name__}` "
                f"initialisation or the `{self.__class__.__name__}.train` method.",
                DeprecationWarning,
                stacklevel=2,
            )

        if trial:  # Trial and model initialization
            self._hp_search_setup(trial)  # sets trainer parameters and initializes model

        args = args or self.args or TrainingArguments()

        if self.train_dataset is None:
            raise ValueError(
                f"Training requires a `train_dataset` given to the `{self.__class__.__name__}` initialization."
            )

        train_parameters = self.dataset_to_parameters(self.train_dataset)
        full_parameters = (
            train_parameters + self.dataset_to_parameters(self.eval_dataset) if self.eval_dataset else train_parameters
        )

        self.train_embeddings(*full_parameters, args=args)
        self.train_classifier(*train_parameters, args=args)

    def dataset_to_parameters(self, dataset: Dataset) -> List[Iterable]:
        return [dataset["text"], dataset["label"]]

    def train_embeddings(
        self,
        x_train: List[str],
        y_train: Optional[Union[List[int], List[List[int]]]] = None,
        x_eval: Optional[List[str]] = None,
        y_eval: Optional[Union[List[int], List[List[int]]]] = None,
        args: Optional[TrainingArguments] = None,
    ) -> None:
        """
        Method to perform the embedding phase: finetuning the `SentenceTransformer` body.

        Args:
            x_train (`List[str]`): A list of training sentences.
            y_train (`Union[List[int], List[List[int]]]`): A list of labels corresponding to the training sentences.
            args (`TrainingArguments`, *optional*):
                Temporarily change the training arguments for this training call.
        """
        if args:
            self.st_trainer.setfit_args = args
        args = args or self.args or TrainingArguments()

        train_max_pairs = -1 if args.max_steps == -1 else args.max_steps * args.embedding_batch_size
        train_dataset, loss = self.get_dataset(x_train, y_train, args=args, max_pairs=train_max_pairs)
        if x_eval is not None and args.eval_strategy != IntervalStrategy.NO:
            eval_max_pairs = -1 if args.eval_max_steps == -1 else args.eval_max_steps * args.embedding_batch_size
            eval_dataset, _ = self.get_dataset(x_eval, y_eval, args=args, max_pairs=eval_max_pairs)
        else:
            eval_dataset = None

        logger.info("***** Running training *****")
        logger.info(f"  Num unique pairs = {len(train_dataset)}")
        logger.info(f"  Batch size = {args.embedding_batch_size}")
        logger.info(f"  Num epochs = {args.embedding_num_epochs}")

        self.st_trainer.train_dataset = train_dataset
        self.st_trainer.eval_dataset = eval_dataset
        self.st_trainer.loss = loss
        if loss in (
            losses.BatchAllTripletLoss,
            losses.BatchHardTripletLoss,
            losses.BatchSemiHardTripletLoss,
            losses.BatchHardSoftMarginTripletLoss,
            SupConLoss,
        ):
            self.st_trainer.args.batch_sampler = BatchSamplers.GROUP_BY_LABEL
        self.st_trainer.train()

    def get_dataset(
        self, x: List[str], y: Union[List[int], List[List[int]]], args: TrainingArguments, max_pairs: int = -1
    ) -> Tuple[Dataset, nn.Module, int, int]:
        if args.loss in [
            losses.BatchAllTripletLoss,
            losses.BatchHardTripletLoss,
            losses.BatchSemiHardTripletLoss,
            losses.BatchHardSoftMarginTripletLoss,
            SupConLoss,
        ]:
            dataset = Dataset.from_dict({"sentence": x, "label": y})

            if args.loss is losses.BatchHardSoftMarginTripletLoss:
                loss = args.loss(
                    model=self.model.model_body,
                    distance_metric=args.distance_metric,
                )
            elif args.loss is SupConLoss:
                loss = args.loss(model=self.model.model_body)
            else:
                loss = args.loss(
                    model=self.model.model_body,
                    distance_metric=args.distance_metric,
                    margin=args.margin,
                )
        else:
            data_sampler = ContrastiveDataset(
                x,
                y,
                self.model.multi_target_strategy,
                args.num_iterations,
                args.sampling_strategy,
                max_pairs=max_pairs,
            )
            dataset = Dataset.from_list(list(data_sampler))
            loss = args.loss(self.model.model_body)

        return dataset, loss

    def _set_logs_prefix(self, logs_prefix: str) -> None:
        """Set the logging prefix.

        Args:
            logs_mapper (str): The logging prefix, e.g. "aspect_embedding".
        """
        self.st_trainer._set_logs_prefix(logs_prefix)

    def train_classifier(
        self, x_train: List[str], y_train: Union[List[int], List[List[int]]], args: Optional[TrainingArguments] = None
    ) -> None:
        """
        Method to perform the classifier phase: fitting a classifier head.

        Args:
            x_train (`List[str]`): A list of training sentences.
            y_train (`Union[List[int], List[List[int]]]`): A list of labels corresponding to the training sentences.
            args (`TrainingArguments`, *optional*):
                Temporarily change the training arguments for this training call.
        """
        args = args or self.args or TrainingArguments()

        self.model.fit(
            x_train,
            y_train,
            num_epochs=args.classifier_num_epochs,
            batch_size=args.classifier_batch_size,
            body_learning_rate=args.body_classifier_learning_rate,
            head_learning_rate=args.head_learning_rate,
            l2_weight=args.l2_weight,
            max_length=args.max_length,
            show_progress_bar=args.show_progress_bar,
            end_to_end=args.end_to_end,
        )

    @torch.no_grad()
    def evaluate(self, dataset: Optional[Dataset] = None, metric_key_prefix: str = "test") -> Dict[str, float]:
        """
        Computes the metrics for a given classifier.

        Args:
            dataset (`Dataset`, *optional*):
                The dataset to compute the metrics on. If not provided, will use the evaluation dataset passed via
                the `eval_dataset` argument at `Trainer` initialization.

        Returns:
            `Dict[str, float]`: The evaluation metrics.
        """

        if dataset is not None:
            self._validate_column_mapping(dataset)
            if self.column_mapping is not None:
                logger.info("Applying column mapping to the evaluation dataset")
                eval_dataset = self._apply_column_mapping(dataset, self.column_mapping)
            else:
                eval_dataset = dataset
        else:
            eval_dataset = self.eval_dataset

        if eval_dataset is None:
            raise ValueError("No evaluation dataset provided to `Trainer.evaluate` nor the `Trainer` initialzation.")

        x_test = eval_dataset["text"]
        y_test = eval_dataset["label"]

        logger.info("***** Running evaluation *****")
        y_pred = self.model.predict(x_test, use_labels=False)
        if isinstance(y_pred, torch.Tensor):
            y_pred = y_pred.cpu()

        # Normalize string outputs
        if y_test and isinstance(y_test[0], str):
            encoder = LabelEncoder()
            encoder.fit(list(y_test) + list(y_pred))
            y_test = encoder.transform(y_test)
            y_pred = encoder.transform(y_pred)

        metric_kwargs = self.metric_kwargs or {}
        if isinstance(self.metric, str):
            metric_config = "multilabel" if self.model.multi_target_strategy is not None else None
            metric_fn = evaluate.load(self.metric, config_name=metric_config)

            results = metric_fn.compute(predictions=y_pred, references=y_test, **metric_kwargs)

        elif callable(self.metric):
            results = self.metric(y_pred, y_test, **metric_kwargs)

        else:
            raise ValueError("metric must be a string or a callable")

        if not isinstance(results, dict):
            results = {"metric": results}
        self.model.model_card_data.post_training_eval_results(
            {f"{metric_key_prefix}_{key}": value for key, value in results.items()}
        )
        return results

    def hyperparameter_search(
        self,
        hp_space: Optional[Callable[["optuna.Trial"], Dict[str, float]]] = None,
        compute_objective: Optional[Callable[[Dict[str, float]], float]] = None,
        n_trials: int = 10,
        direction: str = "maximize",
        backend: Optional[Union["str", HPSearchBackend]] = None,
        hp_name: Optional[Callable[["optuna.Trial"], str]] = None,
        **kwargs,
    ) -> BestRun:
        """
        Launch a hyperparameter search using `optuna`. The optimized quantity is determined
        by `compute_objective`, which defaults to a function returning the evaluation loss when no metric is provided,
        the sum of all metrics otherwise.

        <Tip warning={true}>

        To use this method, you need to have provided a `model_init` when initializing your [`Trainer`]: we need to
        reinitialize the model at each new run.

        </Tip>

        Args:
            hp_space (`Callable[["optuna.Trial"], Dict[str, float]]`, *optional*):
                A function that defines the hyperparameter search space. Will default to
                [`~transformers.trainer_utils.default_hp_space_optuna`].
            compute_objective (`Callable[[Dict[str, float]], float]`, *optional*):
                A function computing the objective to minimize or maximize from the metrics returned by the `evaluate`
                method. Will default to [`~transformers.trainer_utils.default_compute_objective`] which uses the sum of metrics.
            n_trials (`int`, *optional*, defaults to 100):
                The number of trial runs to test.
            direction (`str`, *optional*, defaults to `"maximize"`):
                Whether to optimize greater or lower objects. Can be `"minimize"` or `"maximize"`, you should pick
                `"minimize"` when optimizing the validation loss, `"maximize"` when optimizing one or several metrics.
            backend (`str` or [`~transformers.training_utils.HPSearchBackend`], *optional*):
                The backend to use for hyperparameter search. Only optuna is supported for now.
                TODO: add support for ray and sigopt.
            hp_name (`Callable[["optuna.Trial"], str]]`, *optional*):
                A function that defines the trial/run name. Will default to None.
            kwargs (`Dict[str, Any]`, *optional*):
                Additional keyword arguments passed along to `optuna.create_study`. For more
                information see:

                - the documentation of
                  [optuna.create_study](https://optuna.readthedocs.io/en/stable/reference/generated/optuna.study.create_study.html)

        Returns:
            [`trainer_utils.BestRun`]: All the information about the best run.
        """
        if backend is None:
            backend = default_hp_search_backend()
            if backend is None:
                raise RuntimeError("optuna should be installed. To install optuna run `pip install optuna`.")
        backend = HPSearchBackend(backend)
        if backend == HPSearchBackend.OPTUNA and not is_optuna_available():
            raise RuntimeError("You picked the optuna backend, but it is not installed. Use `pip install optuna`.")
        elif backend != HPSearchBackend.OPTUNA:
            raise RuntimeError("Only optuna backend is supported for hyperparameter search.")
        self.hp_search_backend = backend
        if self.model_init is None:
            raise RuntimeError(
                "To use hyperparameter search, you need to pass your model through a model_init function."
            )

        self.hp_space = default_hp_space_optuna if hp_space is None else hp_space
        self.hp_name = hp_name
        self.compute_objective = default_compute_objective if compute_objective is None else compute_objective

        backend_dict = {
            HPSearchBackend.OPTUNA: run_hp_search_optuna,
        }
        best_run = backend_dict[backend](self, n_trials, direction, **kwargs)

        self.hp_search_backend = None
        return best_run

    def push_to_hub(self, repo_id: str, **kwargs) -> str:
        """Upload model checkpoint to the Hub using `huggingface_hub`.

        See the full list of parameters for your `huggingface_hub` version in the\
        [huggingface_hub documentation](https://huggingface.co/docs/huggingface_hub/package_reference/mixins#huggingface_hub.ModelHubMixin.push_to_hub).

        Args:
            repo_id (`str`):
                The full repository ID to push to, e.g. `"tomaarsen/setfit-sst2"`.
            config (`dict`, *optional*):
                Configuration object to be saved alongside the model weights.
            commit_message (`str`, *optional*):
                Message to commit while pushing.
            private (`bool`, *optional*):
                Whether to make the repo private. If `None` (default), the repo will be public unless the organization's default is private. This value is ignored if the repo already exists.
            api_endpoint (`str`, *optional*):
                The API endpoint to use when pushing the model to the hub.
            token (`str`, *optional*):
                The token to use as HTTP bearer authorization for remote files.
                If not set, will use the token set when logging in with
                `transformers-cli login` (stored in `~/.huggingface`).
            branch (`str`, *optional*):
                The git branch on which to push the model. This defaults to
                the default branch as specified in your repository, which
                defaults to `"main"`.
            create_pr (`boolean`, *optional*):
                Whether or not to create a Pull Request from `branch` with that commit.
                Defaults to `False`.
            allow_patterns (`List[str]` or `str`, *optional*):
                If provided, only files matching at least one pattern are pushed.
            ignore_patterns (`List[str]` or `str`, *optional*):
                If provided, files matching any of the patterns are not pushed.

        Returns:
            str: The url of the commit of your model in the given repository.
        """
        if "/" not in repo_id:
            raise ValueError(
                '`repo_id` must be a full repository ID, including organisation, e.g. "tomaarsen/setfit-sst2".'
            )
        commit_message = kwargs.pop("commit_message", "Add SetFit model")
        return self.model.push_to_hub(repo_id, commit_message=commit_message, **kwargs)


class SetFitTrainer(Trainer):
    """
    `SetFitTrainer` has been deprecated and will be removed in v2.0.0 of SetFit.
    Please use `Trainer` instead.
    """

    def __init__(
        self,
        model: Optional["SetFitModel"] = None,
        train_dataset: Optional["Dataset"] = None,
        eval_dataset: Optional["Dataset"] = None,
        model_init: Optional[Callable[[], "SetFitModel"]] = None,
        metric: Union[str, Callable[["Dataset", "Dataset"], Dict[str, float]]] = "accuracy",
        metric_kwargs: Optional[Dict[str, Any]] = None,
        loss_class=losses.CosineSimilarityLoss,
        num_iterations: int = 20,
        num_epochs: int = 1,
        learning_rate: float = 2e-5,
        batch_size: int = 16,
        seed: int = 42,
        column_mapping: Optional[Dict[str, str]] = None,
        use_amp: bool = False,
        warmup_proportion: float = 0.1,
        distance_metric: Callable = BatchHardTripletLossDistanceFunction.cosine_distance,
        margin: float = 0.25,
        samples_per_label: int = 2,
    ):
        warnings.warn(
            "`SetFitTrainer` has been deprecated and will be removed in v2.0.0 of SetFit. "
            "Please use `Trainer` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        args = TrainingArguments(
            num_iterations=num_iterations,
            num_epochs=num_epochs,
            body_learning_rate=learning_rate,
            head_learning_rate=learning_rate,
            batch_size=batch_size,
            seed=seed,
            use_amp=use_amp,
            warmup_proportion=warmup_proportion,
            distance_metric=distance_metric,
            margin=margin,
            samples_per_label=samples_per_label,
            loss=loss_class,
        )
        super().__init__(
            model=model,
            args=args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            model_init=model_init,
            metric=metric,
            metric_kwargs=metric_kwargs,
            column_mapping=column_mapping,
        )



import numpy as np
import pytest

from setfit.sampler import ContrastiveDataset


@pytest.mark.parametrize(
    "sampling_strategy, expected_pos_pairs, expected_neg_pairs",
    [("unique", 4, 2), ("undersampling", 2, 2), ("oversampling", 4, 4)],
)
def test_sentence_pairs_generation(sampling_strategy: str, expected_pos_pairs: int, expected_neg_pairs: int):
    sentences = np.array(["sent 1", "sent 2", "sent 3"])
    labels = np.array(["label 1", "label 1", "label 2"])

    multilabel = False

    data_sampler = ContrastiveDataset(sentences, labels, multilabel, sampling_strategy=sampling_strategy)

    assert data_sampler.len_pos_pairs == expected_pos_pairs
    assert data_sampler.len_neg_pairs == expected_neg_pairs

    pairs = [i for i in data_sampler]

    assert len(pairs) == expected_pos_pairs + expected_neg_pairs
    assert pairs[0] == {"sentence_1": "sent 1", "sentence_2": "sent 1", "label": 1.0}


@pytest.mark.parametrize(
    "sampling_strategy, expected_pos_pairs, expected_neg_pairs",
    [("unique", 6, 4), ("undersampling", 4, 4), ("oversampling", 6, 6)],
)
def test_sentence_pairs_generation_multilabel(
    sampling_strategy: str, expected_pos_pairs: int, expected_neg_pairs: int
):
    sentences = np.array(["sent 1", "sent 2", "sent 3", "sent 4"])
    labels = np.array([[1, 0, 0, 1], [0, 1, 1, 0], [0, 0, 1, 0], [0, 0, 0, 1]])

    multilabel = True

    data_sampler = ContrastiveDataset(sentences, labels, multilabel, sampling_strategy=sampling_strategy)
    assert data_sampler.len_pos_pairs == expected_pos_pairs
    assert data_sampler.len_neg_pairs == expected_neg_pairs

    pairs = [i for i in data_sampler]
    assert len(pairs) == expected_pos_pairs + expected_neg_pairs



import re

from transformers.utils.notebook import NotebookProgressCallback


class SetFitNotebookProgressCallback(NotebookProgressCallback):
    """
    A variation of NotebookProgressCallback that accepts logs/metrics other than "loss" and "eval_loss".
    In particular, it accepts "embedding_loss", "aspect_embedding_loss", and "polarity_embedding_loss"
    and the corresponding metrics for the validation set.
    """

    def on_log(self, *args, logs=None, **kwargs):
        if logs is not None:
            logs = {key if key != "embedding_loss" else "loss": value for key, value in logs.items()}
        return super().on_log(*args, logs=logs, **kwargs)

    def on_evaluate(self, args, state, control, metrics=None, **kwargs):
        if self.training_tracker is not None:
            values = {"Training Loss": "No log", "Validation Loss": "No log"}
            for log in reversed(state.log_history):
                if loss_logs := {
                    key for key in log if key in ("embedding_loss", "aspect_embedding_loss", "polarity_embedding_loss")
                }:
                    values["Training Loss"] = log[loss_logs.pop()]
                    break

            if self.first_column == "Epoch":
                values["Epoch"] = int(state.epoch)
            else:
                values["Step"] = state.global_step
            metric_key_prefix = "eval"
            for k in metrics:
                if k.endswith("_loss"):
                    metric_key_prefix = re.sub(r"\_loss$", "", k)
            _ = metrics.pop("total_flos", None)
            _ = metrics.pop("epoch", None)
            _ = metrics.pop(f"{metric_key_prefix}_runtime", None)
            _ = metrics.pop(f"{metric_key_prefix}_samples_per_second", None)
            _ = metrics.pop(f"{metric_key_prefix}_steps_per_second", None)
            _ = metrics.pop(f"{metric_key_prefix}_jit_compilation_time", None)
            for k, v in metrics.items():
                splits = k.split("_")
                name = " ".join([part.capitalize() for part in splits[1:]])
                if name in ("Embedding Loss", "Aspect Embedding Loss", "Polarity Embedding Loss"):
                    # Single dataset
                    name = "Validation Loss"
                values[name] = v
            self.training_tracker.write_line(values)
            self.training_tracker.remove_child()
            self.prediction_bar = None
            # Evaluation takes a long time so we should force the next update.
            self._force_next_update = True



import string

import numpy as np
import pandas as pd
import pytest
from datasets import Dataset, load_dataset
from torch.utils.data import DataLoader
from transformers import AutoTokenizer

from setfit.data import (
    SAMPLE_SIZES,
    SEEDS,
    SetFitDataset,
    create_fewshot_splits,
    create_fewshot_splits_multilabel,
    create_samples,
    get_templated_dataset,
    sample_dataset,
)


@pytest.fixture
def empty_dataset():
    return Dataset.from_dict({})


@pytest.fixture
def dataset():
    return Dataset.from_dict(
        {
            "text": ["label-0 text", "label-1 text"],
            "label": [[1, 0], [0, 1]],
        }
    )


@pytest.fixture
def unbalanced_dataset():
    return Dataset.from_dict({"text": string.ascii_letters, "label": [0] + 51 * [1]})


def test_add_to_empty_dataset_defaults(empty_dataset):
    augmented_dataset = get_templated_dataset(empty_dataset, candidate_labels=["label-0", "label-1"], multi_label=True)

    assert augmented_dataset[:] == {
        "text": [
            "This sentence is label-0",
            "This sentence is label-0",
            "This sentence is label-1",
            "This sentence is label-1",
        ],
        "label": [[1, 0], [1, 0], [0, 1], [0, 1]],
    }


def test_add_to_dataset_defaults(dataset):
    augmented_dataset = get_templated_dataset(dataset, candidate_labels=["label-0", "label-1"], multi_label=True)

    assert augmented_dataset[:] == {
        "text": [
            "label-0 text",
            "label-1 text",
            "This sentence is label-0",
            "This sentence is label-0",
            "This sentence is label-1",
            "This sentence is label-1",
        ],
        "label": [[1, 0], [0, 1], [1, 0], [1, 0], [0, 1], [0, 1]],
    }


@pytest.mark.parametrize(
    "text_column, label_column",
    [
        ("missing-text", "label"),
        ("text", "missing-label"),
        ("missing-text", "missing-label"),
    ],
)
def test_missing_columns(dataset, text_column, label_column):
    with pytest.raises(ValueError):
        get_templated_dataset(
            dataset,
            candidate_labels=["label-0", "label-1"],
            text_column=text_column,
            label_column=label_column,
        )


@pytest.mark.parametrize("sample_size", [8, 16, 32])
def test_subset_is_larger_than_sample_size(sample_size):
    data = {"label": [0] * 50 + [1] * 50}
    df = pd.DataFrame(data)
    sample_df = create_samples(df, sample_size=sample_size, seed=0)
    assert len(sample_df) == (sample_size * 2)


@pytest.mark.parametrize("sample_size", [8, 16, 32])
def test_subset_is_smaller_than_sample_size(sample_size):
    data = {"label": [0] * 3 + [1] * 3}
    df = pd.DataFrame(data)
    sample_df = create_samples(df, sample_size=sample_size, seed=0)
    assert len(sample_df) == len(df)


def test_expected_number_of_splits():
    dataset = Dataset.from_pandas(pd.DataFrame({"label": [0] * 50 + [1] * 50}))
    num_labels = 2
    splits_ds = create_fewshot_splits(dataset, SAMPLE_SIZES)
    assert len(splits_ds) == len(SAMPLE_SIZES) * len(SEEDS)

    split: Dataset
    for idx, split in enumerate(splits_ds.values()):
        sample_size = SAMPLE_SIZES[idx // len(SEEDS)]
        # The number of rows is limited by 100 due to the size of the original dataset
        assert len(split) == min(sample_size * num_labels, len(dataset))


def test_create_fewshot_splits_with_augmentation():
    dataset_name = "sst5"
    dataset = load_dataset(f"SetFit/{dataset_name}", split="train")
    num_labels = len(set(dataset["label"]))
    splits_ds = create_fewshot_splits(
        dataset, SAMPLE_SIZES, add_data_augmentation=True, dataset_name=f"SetFit/{dataset_name}"
    )
    assert len(splits_ds) == len(SAMPLE_SIZES) * len(SEEDS)

    split: Dataset
    for idx, split in enumerate(splits_ds.values()):
        sample_size = SAMPLE_SIZES[idx // len(SEEDS)]
        # Each split should have sample_size * num_labels * 2 rows:
        # for each label we sample `sample_size`, and then we generate
        # another `sample_size` samples through augmentation.
        assert len(split) == sample_size * num_labels * 2


def test_create_fewshot_splits_multilabel():
    num_samples = 50
    dataset = Dataset.from_dict(
        {
            "text": string.ascii_letters[:50],
            "label_one": np.random.randint(2, size=(num_samples,)),
            "label_two": np.random.randint(2, size=(num_samples,)),
            "label_three": np.random.randint(2, size=(num_samples,)),
        }
    )
    splits_ds = create_fewshot_splits_multilabel(dataset, SAMPLE_SIZES)
    assert len(splits_ds) == len(SAMPLE_SIZES) * len(SEEDS)
    # We can't safely test the number of rows of each of the splits
    # as duplicate samples are removed.


def test_sample_dataset_returns_expected_samples():
    num_samples = 2
    dataset = Dataset.from_dict({"text": ["hello"] * 50, "label": [0] * 25 + [1] * 25})
    samples = sample_dataset(dataset=dataset, num_samples=num_samples)
    for label_id in range(num_samples):
        assert len(samples.filter(lambda x: x["label"] == label_id)) == num_samples


def test_sample_dataset_with_label_column():
    num_samples = 2
    label_column = "my_labels"
    dataset = Dataset.from_dict({"text": ["hello"] * 50, label_column: [0] * 25 + [1] * 25})
    samples = sample_dataset(dataset=dataset, label_column=label_column, num_samples=num_samples)
    for label_id in range(num_samples):
        assert len(samples.filter(lambda x: x[label_column] == label_id)) == num_samples


def test_sample_dataset_with_unbalanced_ds(unbalanced_dataset):
    num_samples = 8
    ds = sample_dataset(unbalanced_dataset, num_samples=num_samples)
    # The dataset ought to have just `num_samples + 1` rows, as `unbalanced_dataset`
    # has one label with more than `num_samples` entries and another label with just 1 row.
    # We sample `num_samples` from the former, and 1 from the latter.
    assert ds.num_rows == num_samples + 1


@pytest.mark.parametrize(
    "dataset",
    [
        "SetFit/emotion",
        "SetFit/ag_news",
        "SetFit/amazon_counterfactual_en",
        "SetFit/SentEval-CR",
        "SetFit/sst5",
        "SetFit/enron_spam",
        "SetFit/tweet_eval_stance_abortion",
        "SetFit/ade_corpus_v2_classification",
    ],
)
def test_get_augmented_samples(dataset: str):
    dataset = get_templated_dataset(reference_dataset=dataset)
    assert set(dataset.column_names) == {"text", "label"}
    assert len(dataset["text"])
    assert len(dataset["label"])


def test_get_augmented_samples_negative():
    with pytest.raises(ValueError):
        get_templated_dataset(reference_dataset=None, candidate_labels=None)


@pytest.mark.parametrize(
    "tokenizer_name",
    ["sentence-transformers/paraphrase-albert-small-v2", "sentence-transformers/distiluse-base-multilingual-cased-v1"],
)
def test_correct_model_inputs(tokenizer_name):
    # Arbitrary testing data
    x = list(string.ascii_lowercase)
    y = list(range(len(x)))

    # Relatively Standard DataLoader setup using a SetFitDataset
    # for training a differentiable classification head
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
    dataset = SetFitDataset(x, y, tokenizer)
    dataloader = DataLoader(
        dataset,
        batch_size=2,
        collate_fn=dataset.collate_fn,
        shuffle=True,
        pin_memory=True,
    )

    # Verify that the x_batch contains exactly those keys that the model requires
    x_batch, _ = next(iter(dataloader))
    assert set(x_batch.keys()) == set(tokenizer.model_input_names)


def test_preserve_features() -> None:
    dataset = load_dataset("SetFit/sst5", split="train[:100]")
    label_column = "label_text"
    dataset = dataset.class_encode_column(label_column)
    train_dataset = sample_dataset(dataset, label_column=label_column, num_samples=8)
    assert train_dataset.features[label_column] == dataset.features[label_column]



from pathlib import Path

import datasets
import pytest
from datasets import Dataset, load_dataset
from packaging.version import Version, parse

from setfit import SetFitModel, SetFitModelCardData, Trainer, TrainingArguments
from setfit.data import sample_dataset
from setfit.model_card import generate_model_card, is_on_huggingface

from .model_card_pattern import MODEL_CARD_PATTERN


def test_model_card(tmp_path: Path) -> None:
    dataset = load_dataset("sst2")
    train_dataset = sample_dataset(dataset["train"], label_column="label", num_samples=8)
    eval_dataset = dataset["validation"].select(range(10))
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2",
        labels=["negative", "positive"],
        model_card_data=SetFitModelCardData(
            model_id="tomaarsen/setfit-paraphrase-albert-small-v2-sst2",
            dataset_id="sst2",
            dataset_name="SST2",
            language=["en"],
            license="apache-2.0",
        ),
    )

    args = TrainingArguments(
        str(tmp_path),
        report_to="codecarbon",
        batch_size=1,
        eval_steps=1,
        logging_steps=1,
        max_steps=2,
        eval_strategy="steps",
        save_strategy="no",
    )
    trainer = Trainer(
        model=model,
        args=args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        column_mapping={"sentence": "text"},
    )
    trainer.train()
    trainer.evaluate()
    model_card = generate_model_card(trainer.model)
    assert MODEL_CARD_PATTERN.fullmatch(model_card)


def test_model_card_languages() -> None:
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2",
        model_card_data=SetFitModelCardData(
            language=["en", "nl", "de"],
        ),
    )
    model_card = model.generate_model_card()
    assert "**Languages:** en, nl, de" in model_card


def test_is_on_huggingface_edge_case() -> None:
    assert not is_on_huggingface("test_value")
    assert not is_on_huggingface("a/test/value")


@pytest.mark.skipif(
    parse(datasets.__version__) < Version("2.14.0"), reason="Inferring dataset_id only works from datasets >= 2.14.0"
)
@pytest.mark.parametrize("dataset_id", ("SetFit/emotion", "SetFit/sst2"))
def test_infer_dataset_id(dataset_id: str) -> None:
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    train_dataset = load_dataset(dataset_id, split="train")

    # This triggers inferring the dataset_id from train_dataset
    Trainer(model=model, train_dataset=train_dataset)
    assert model.model_card_data.dataset_id == dataset_id


def test_cant_infer_dataset_id():
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    train_dataset = Dataset.from_dict({"text": ["a", "b", "c", "d"], "label": [0, 1, 1, 0]})

    # This triggers inferring the dataset_id from train_dataset
    Trainer(model=model, train_dataset=train_dataset)
    assert model.model_card_data.dataset_id is None



from unittest import TestCase

import pytest
from datasets import Dataset

from setfit import DistillationTrainer, Trainer
from setfit.modeling import SetFitModel
from setfit.training_args import TrainingArguments


class DistillationTrainerTest(TestCase):
    def setUp(self):
        self.teacher_model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
        self.student_model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-MiniLM-L3-v2")
        self.args = TrainingArguments(num_iterations=1)

    def test_trainer_works_with_default_columns(self):
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        # train a teacher model
        teacher_trainer = Trainer(
            model=self.teacher_model,
            train_dataset=dataset,
            eval_dataset=dataset,
            metric="accuracy",
        )
        # Teacher Train and evaluate
        teacher_trainer.train()
        teacher_model = teacher_trainer.model

        student_trainer = DistillationTrainer(
            teacher_model=teacher_model,
            train_dataset=dataset,
            student_model=self.student_model,
            eval_dataset=dataset,
            metric="accuracy",
        )

        # Student Train and evaluate
        student_trainer.train()
        metrics = student_trainer.evaluate()
        print("Student results: ", metrics)
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_raises_error_with_missing_label(self):
        labeled_dataset = Dataset.from_dict(
            {"text": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        # train a teacher model
        teacher_trainer = Trainer(
            model=self.teacher_model,
            train_dataset=labeled_dataset,
            eval_dataset=labeled_dataset,
            metric="accuracy",
            args=self.args,
        )
        # Teacher Train and evaluate
        teacher_trainer.train()

        unlabeled_dataset = Dataset.from_dict({"text": ["a", "b", "c"], "extra_column": ["d", "e", "f"]})
        student_trainer = DistillationTrainer(
            teacher_model=self.teacher_model,
            student_model=self.student_model,
            train_dataset=unlabeled_dataset,
            eval_dataset=labeled_dataset,
            args=self.args,
        )
        student_trainer.train()
        metrics = student_trainer.evaluate()
        print("Student results: ", metrics)
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_raises_error_with_missing_text(self):
        dataset = Dataset.from_dict({"label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        with pytest.raises(ValueError):
            DistillationTrainer(
                teacher_model=self.teacher_model,
                train_dataset=dataset,
                student_model=self.student_model,
                eval_dataset=dataset,
                args=self.args,
            )

    def test_column_mapping_with_missing_text(self):
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "extra_column": ["d", "e", "f"]})
        with pytest.raises(ValueError):
            DistillationTrainer(
                teacher_model=self.teacher_model,
                train_dataset=dataset,
                student_model=self.student_model,
                eval_dataset=dataset,
                args=self.args,
                column_mapping={"label_new": "label"},
            )

    def test_column_mapping_multilabel(self):
        dataset = Dataset.from_dict({"text_new": ["a", "b", "c"], "label_new": [[0, 1], [1, 2], [2, 0]]})

        trainer = DistillationTrainer(
            teacher_model=self.teacher_model,
            train_dataset=dataset,
            student_model=self.student_model,
            eval_dataset=dataset,
            args=self.args,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer._validate_column_mapping(dataset)
        formatted_dataset = trainer._apply_column_mapping(dataset, trainer.column_mapping)

        assert formatted_dataset.column_names == ["text", "label"]
        assert formatted_dataset[0]["text"] == "a"
        assert formatted_dataset[0]["label"] == [0, 1]
        assert formatted_dataset[1]["text"] == "b"


@pytest.mark.parametrize("teacher_diff", [True, False])
@pytest.mark.parametrize("student_diff", [True, False])
def test_differentiable_models(teacher_diff: bool, student_diff: bool) -> None:
    if teacher_diff:
        teacher_model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2",
            use_differentiable_head=True,
            head_params={"out_features": 3},
        )
    else:
        teacher_model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    if student_diff:
        student_model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-MiniLM-L3-v2",
            use_differentiable_head=True,
            head_params={"out_features": 3},
        )
    else:
        student_model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-MiniLM-L3-v2")

    dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
    # train a teacher model
    teacher_trainer = Trainer(
        model=teacher_model,
        train_dataset=dataset,
        eval_dataset=dataset,
        metric="accuracy",
    )
    teacher_trainer.train()
    metrics = teacher_trainer.evaluate()
    print("Teacher results: ", metrics)
    assert metrics["accuracy"] == 1.0
    teacher_model = teacher_trainer.model

    student_trainer = DistillationTrainer(
        teacher_model=teacher_model,
        train_dataset=dataset,
        student_model=student_model,
        eval_dataset=dataset,
        metric="accuracy",
    )

    # Student Train and evaluate
    student_trainer.train()
    metrics = student_trainer.evaluate()
    print("Student results: ", metrics)
    assert metrics["accuracy"] == 1.0



import os
import re
from pathlib import Path
from unittest import TestCase

import evaluate
import pytest
import torch
from datasets import Dataset, load_dataset
from sentence_transformers import losses
from transformers import TrainerCallback
from transformers.testing_utils import require_optuna
from transformers.utils.hp_naming import TrialShortNamer

from setfit import logging
from setfit.losses import SupConLoss
from setfit.modeling import SetFitModel
from setfit.trainer import Trainer
from setfit.training_args import TrainingArguments
from setfit.utils import BestRun
from tests.utils import SafeTemporaryDirectory


logging.set_verbosity_warning()
logging.enable_propagation()


class TrainerTest(TestCase):
    def setUp(self):
        self.model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
        self.args = TrainingArguments(num_iterations=1)

    def test_trainer_works_with_model_init(self):
        def get_model():
            model_name = "sentence-transformers/paraphrase-albert-small-v2"
            return SetFitModel.from_pretrained(model_name)

        dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        trainer = Trainer(
            model_init=get_model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )
        trainer.train()
        metrics = trainer.evaluate()
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_works_with_column_mapping(self):
        dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )
        trainer.train()
        metrics = trainer.evaluate()
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_works_with_partial_column_mapping(self):
        dataset = Dataset.from_dict({"text_new": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            column_mapping={"text_new": "text"},
        )
        trainer.train()
        metrics = trainer.evaluate()
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_works_with_default_columns(self):
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        trainer = Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)
        trainer.train()
        metrics = trainer.evaluate()
        self.assertEqual(metrics["accuracy"], 1.0)

    def test_trainer_works_with_alternate_dataset_for_evaluate(self):
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        alternate_dataset = Dataset.from_dict(
            {"text": ["x", "y", "z"], "label": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        trainer = Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)
        trainer.train()
        metrics = trainer.evaluate(alternate_dataset)
        self.assertNotEqual(metrics["accuracy"], 1.0)

    def test_trainer_raises_error_with_missing_label(self):
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "extra_column": ["d", "e", "f"]})
        with pytest.raises(ValueError):
            Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)

    def test_trainer_raises_error_with_missing_text(self):
        """If the required columns are missing from the dataset, the library should throw an error and list the columns found."""
        dataset = Dataset.from_dict({"label": [0, 1, 2], "extra_column": ["d", "e", "f"]})
        expected_message = re.escape(
            "SetFit expected the dataset to have the columns ['label', 'text'], "
            "but only the columns ['extra_column', 'label'] were found. "
            "Either make sure these columns are present, or specify which columns to use with column_mapping in Trainer."
        )
        with pytest.raises(ValueError, match=expected_message):
            Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)

    def test_column_mapping_raises_error_when_mapped_columns_missing(self):
        """If the columns specified in the column mapping are missing from the dataset, the library should throw an error and list the columns found."""
        dataset = Dataset.from_dict({"text": ["a", "b", "c"], "extra_column": ["d", "e", "f"]})
        expected_message = re.escape(
            "The column mapping expected the columns ['label_new', 'text_new'] in the dataset, "
            "but the dataset had the columns ['extra_column', 'text'].",
        )
        with pytest.raises(ValueError, match=expected_message):
            Trainer(
                model=self.model,
                args=self.args,
                train_dataset=dataset,
                eval_dataset=dataset,
                column_mapping={"text_new": "text", "label_new": "label"},
            )

    def test_trainer_raises_error_when_dataset_not_split(self):
        """Verify that an error is raised if we pass an unsplit dataset to the trainer."""
        dataset = Dataset.from_dict({"text": ["a", "b", "c", "d"], "label": [0, 0, 1, 1]}).train_test_split(
            test_size=0.5
        )
        expected_message = re.escape(
            "SetFit expected a Dataset, but it got a DatasetDict with the splits ['test', 'train']. "
            "Did you mean to select one of these splits from the dataset?",
        )
        with pytest.raises(ValueError, match=expected_message):
            Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)

    def test_trainer_raises_error_when_dataset_is_dataset_dict_with_train(self):
        """Verify that a useful error is raised if we pass an unsplit dataset with only a `train` split to the trainer."""
        with SafeTemporaryDirectory() as tmpdirname:
            path = Path(tmpdirname) / "test_dataset_dict_with_train.csv"
            path.write_text("label,text\n1,good\n0,terrible\n")
            dataset = load_dataset("csv", data_files=str(path))
        expected_message = re.escape(
            "SetFit expected a Dataset, but it got a DatasetDict with the split ['train']. "
            "Did you mean to select the training split with dataset['train']?",
        )
        with pytest.raises(ValueError, match=expected_message):
            Trainer(model=self.model, args=self.args, train_dataset=dataset, eval_dataset=dataset)

    def test_column_mapping_multilabel(self):
        dataset = Dataset.from_dict({"text_new": ["a", "b", "c"], "label_new": [[0, 1], [1, 2], [2, 0]]})

        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer._validate_column_mapping(dataset)
        formatted_dataset = trainer._apply_column_mapping(dataset, trainer.column_mapping)

        assert formatted_dataset.column_names == ["text", "label"]

        assert formatted_dataset[0]["text"] == "a"
        assert formatted_dataset[0]["label"] == [0, 1]

        assert formatted_dataset[1]["text"] == "b"

    def test_trainer_support_callable_as_metric(self):
        dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )

        f1_metric = evaluate.load("f1")
        accuracy_metric = evaluate.load("accuracy")

        def compute_metrics(y_pred, y_test):
            return {
                "f1": f1_metric.compute(predictions=y_pred, references=y_test, average="micro")["f1"],
                "accuracy": accuracy_metric.compute(predictions=y_pred, references=y_test)["accuracy"],
            }

        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            metric=compute_metrics,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer.train()
        metrics = trainer.evaluate()

        self.assertEqual(
            {
                "f1": 1.0,
                "accuracy": 1.0,
            },
            metrics,
        )

    def test_raise_when_metric_value_is_invalid(self):
        dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )

        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            metric="this-metric-does-not-exist",  # invalid metric value
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer.train()

        with self.assertRaises(FileNotFoundError):
            trainer.evaluate()


class TrainerDifferentiableHeadTest(TestCase):
    def setUp(self):
        self.dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        self.model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2",
            use_differentiable_head=True,
            head_params={"out_features": 3},
        )
        self.args = TrainingArguments(num_iterations=1)

    def test_trainer_normalize(self):
        self.model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2",
            use_differentiable_head=True,
            head_params={"out_features": 3},
            normalize_embeddings=True,
        )
        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=self.dataset,
            eval_dataset=self.dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )
        trainer.train()
        metrics = trainer.evaluate()
        self.assertEqual(metrics, {"accuracy": 1.0})

    @pytest.mark.skip(
        "This test is flaky and needs to be fixed; it passes when run in isolation, under "
        "TrainerDifferentiableHeadTest, and even under test_trainer.py, but fails when run in the full test suite."
    )
    def test_trainer_max_length_exceeds_max_acceptable_length(self):
        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=self.dataset,
            eval_dataset=self.dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )
        trainer.unfreeze(keep_body_frozen=True)
        with self.assertLogs(level=logging.WARNING) as cm:
            max_length = 4096
            max_acceptable_length = self.model.model_body.get_max_seq_length()
            args = TrainingArguments(num_iterations=1, max_length=max_length)
            trainer.train(args)
            self.assertEqual(
                cm.output,
                [
                    (
                        f"WARNING:setfit.modeling:The specified `max_length`: {max_length} is greater than the maximum length "
                        f"of the current model body: {max_acceptable_length}. Using {max_acceptable_length} instead."
                    )
                ],
            )

    def test_trainer_max_length_is_smaller_than_max_acceptable_length(self):
        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=self.dataset,
            eval_dataset=self.dataset,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        # An alternative way of `assertNoLogs`, which is new in Python 3.10
        try:
            with self.assertLogs(level=logging.WARNING) as cm:
                max_length = 32
                args = TrainingArguments(num_iterations=1, max_length=max_length)
                trainer.train(args)
                self.assertEqual(cm.output, [])
        except AssertionError as e:
            if e.args[0] != "no logs of level WARNING or higher triggered on root":
                raise AssertionError(e)


class TrainerMultilabelTest(TestCase):
    def setUp(self):
        self.model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2", multi_target_strategy="one-vs-rest"
        )
        self.args = TrainingArguments(num_iterations=1)

    def test_trainer_multilabel_support_callable_as_metric(self):
        dataset = Dataset.from_dict({"text_new": ["a", "b", "c"], "label_new": [[1, 0, 0], [0, 1, 0], [0, 0, 1]]})

        multilabel_f1_metric = evaluate.load("f1", "multilabel")
        multilabel_accuracy_metric = evaluate.load("accuracy", "multilabel")

        def compute_metrics(y_pred, y_test):
            return {
                "f1": multilabel_f1_metric.compute(predictions=y_pred, references=y_test, average="micro")["f1"],
                "accuracy": multilabel_accuracy_metric.compute(predictions=y_pred, references=y_test)["accuracy"],
            }

        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            metric=compute_metrics,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer.train()
        metrics = trainer.evaluate()

        self.assertEqual(
            {
                "f1": 1.0,
                "accuracy": 1.0,
            },
            metrics,
        )


class TrainerMultilabelDifferentiableTest(TestCase):
    def setUp(self):
        self.model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2",
            multi_target_strategy="one-vs-rest",
            use_differentiable_head=True,
            head_params={"out_features": 2},
        )
        self.args = TrainingArguments(num_iterations=1)

    def test_trainer_multilabel_support_callable_as_metric(self):
        dataset = Dataset.from_dict({"text_new": ["", "a", "b", "ab"], "label_new": [[0, 0], [1, 0], [0, 1], [1, 1]]})

        multilabel_f1_metric = evaluate.load("f1", "multilabel")
        multilabel_accuracy_metric = evaluate.load("accuracy", "multilabel")

        def compute_metrics(y_pred, y_test):
            return {
                "f1": multilabel_f1_metric.compute(predictions=y_pred, references=y_test, average="micro")["f1"],
                "accuracy": multilabel_accuracy_metric.compute(predictions=y_pred, references=y_test)["accuracy"],
            }

        trainer = Trainer(
            model=self.model,
            args=self.args,
            train_dataset=dataset,
            eval_dataset=dataset,
            metric=compute_metrics,
            column_mapping={"text_new": "text", "label_new": "label"},
        )

        trainer.train()
        metrics = trainer.evaluate()

        self.assertEqual(
            {
                "f1": 1.0,
                "accuracy": 1.0,
            },
            metrics,
        )


@require_optuna
class TrainerHyperParameterOptunaIntegrationTest(TestCase):
    def setUp(self):
        self.dataset = Dataset.from_dict(
            {"text_new": ["a", "b", "c"], "label_new": [0, 1, 2], "extra_column": ["d", "e", "f"]}
        )
        self.args = TrainingArguments(num_iterations=1)

    def test_hyperparameter_search(self):
        class MyTrialShortNamer(TrialShortNamer):
            DEFAULTS = {"max_iter": 100, "solver": "liblinear"}

        def hp_space(trial):
            return {
                "learning_rate": trial.suggest_float("learning_rate", 1e-6, 1e-4, log=True),
                "batch_size": trial.suggest_categorical("batch_size", [4, 8, 16, 32, 64]),
                "max_iter": trial.suggest_int("max_iter", 50, 300),
                "solver": trial.suggest_categorical("solver", ["newton-cg", "lbfgs", "liblinear"]),
            }

        def model_init(params):
            params = params or {}
            max_iter = params.get("max_iter", 100)
            solver = params.get("solver", "liblinear")
            params = {
                "head_params": {
                    "max_iter": max_iter,
                    "solver": solver,
                }
            }
            return SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2", **params)

        def hp_name(trial):
            return MyTrialShortNamer.shortname(trial.params)

        trainer = Trainer(
            args=self.args,
            train_dataset=self.dataset,
            eval_dataset=self.dataset,
            model_init=model_init,
            column_mapping={"text_new": "text", "label_new": "label"},
        )
        result = trainer.hyperparameter_search(direction="minimize", hp_space=hp_space, hp_name=hp_name, n_trials=4)
        assert isinstance(result, BestRun)
        assert result.hyperparameters.keys() == {"learning_rate", "batch_size", "max_iter", "solver"}


# regression test for https://github.com/huggingface/setfit/issues/153
@pytest.mark.parametrize(
    "loss_class",
    [
        losses.BatchAllTripletLoss,
        losses.BatchHardTripletLoss,
        losses.BatchSemiHardTripletLoss,
        losses.BatchHardSoftMarginTripletLoss,
        SupConLoss,
    ],
)
def test_trainer_works_with_non_default_loss_class(loss_class):
    dataset = Dataset.from_dict({"text": ["a 1", "b 1", "c 1", "a 2", "b 2", "c 2"], "label": [0, 1, 2, 0, 1, 2]})
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    args = TrainingArguments(num_iterations=1, loss=loss_class)
    trainer = Trainer(
        model=model,
        args=args,
        train_dataset=dataset,
        eval_dataset=dataset,
    )
    trainer.train()
    # no asserts here because this is a regression test - we only test if an exception is raised


def test_trainer_evaluate_with_strings(model: SetFitModel):
    dataset = Dataset.from_dict(
        {"text": ["positive sentence", "negative sentence"], "label": ["positive", "negative"]}
    )
    trainer = Trainer(
        model=model,
        args=TrainingArguments(num_iterations=1),
        train_dataset=dataset,
        eval_dataset=dataset,
    )
    trainer.train()
    # This used to fail due to "TypeError: can't convert np.ndarray of type numpy.str_.
    # The only supported types are: float64, float32, float16, complex64, complex128, int64, int32, int16, int8, uint8, and bool."
    model.predict(["another positive sentence"])

    assert set(model.labels) == {"positive", "negative"}


def test_trainer_evaluate_multilabel_f1():
    dataset = Dataset.from_dict({"text_new": ["", "a", "b", "ab"], "label_new": [[0, 0], [1, 0], [0, 1], [1, 1]]})
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", multi_target_strategy="one-vs-rest"
    )

    trainer = Trainer(
        model=model,
        args=TrainingArguments(num_iterations=5),
        train_dataset=dataset,
        eval_dataset=dataset,
        metric="f1",
        metric_kwargs={"average": "micro"},
        column_mapping={"text_new": "text", "label_new": "label"},
    )

    trainer.train()
    metrics = trainer.evaluate()
    assert metrics == {"f1": 1.0}


def test_trainer_evaluate_on_cpu() -> None:
    # This test used to fail if CUDA was available
    dataset = Dataset.from_dict({"text": ["positive sentence", "negative sentence"], "label": [1, 0]})
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", use_differentiable_head=True
    )

    def compute_metric(y_pred, y_test) -> None:
        assert y_pred.device == torch.device("cpu")
        return 1.0

    args = TrainingArguments(num_iterations=5)
    trainer = Trainer(
        model=model,
        args=args,
        train_dataset=dataset,
        eval_dataset=dataset,
        metric=compute_metric,
    )
    trainer.train()
    trainer.evaluate()


def test_no_model_no_model_init():
    with pytest.raises(RuntimeError, match="`Trainer` requires either a `model` or `model_init` argument."):
        Trainer()


def test_model_and_model_init(model: SetFitModel):
    def model_init() -> SetFitModel:
        return model

    with pytest.raises(RuntimeError, match="`Trainer` requires either a `model` or `model_init` argument."):
        Trainer(model=model, model_init=model_init)


def test_trainer_callbacks(model: SetFitModel):
    trainer = Trainer(model=model)
    assert len(trainer.st_trainer.callback_handler.callbacks) >= 2
    num_callbacks = len(trainer.st_trainer.callback_handler.callbacks)
    callback_names = {callback.__class__.__name__ for callback in trainer.st_trainer.callback_handler.callbacks}
    assert {"DefaultFlowCallback", "ProgressCallback"} <= callback_names

    class TestCallback(TrainerCallback):
        pass

    callback = TestCallback()
    trainer.add_callback(callback)
    assert len(trainer.st_trainer.callback_handler.callbacks) == num_callbacks + 1
    assert trainer.st_trainer.callback_handler.callbacks[-1] == callback

    assert trainer.pop_callback(callback) == callback
    trainer.add_callback(callback)
    assert trainer.st_trainer.callback_handler.callbacks[-1] == callback
    trainer.remove_callback(callback)
    assert callback not in trainer.st_trainer.callback_handler.callbacks


def test_trainer_warn_freeze(model: SetFitModel):
    trainer = Trainer(model)
    with pytest.warns(
        DeprecationWarning,
        match="Trainer.freeze` is deprecated and will be removed in v2.0.0 of SetFit. "
        "Please use `SetFitModel.freeze` directly instead.",
    ):
        trainer.freeze()


def test_train_with_kwargs(model: SetFitModel) -> None:
    train_dataset = Dataset.from_dict({"text": ["positive sentence", "negative sentence"], "label": [1, 0]})
    trainer = Trainer(model, train_dataset=train_dataset)
    with pytest.warns(DeprecationWarning, match="`Trainer.train` does not accept keyword arguments anymore."):
        trainer.train(num_epochs=5)


def test_train_no_dataset(model: SetFitModel) -> None:
    trainer = Trainer(model)
    with pytest.raises(ValueError, match="Training requires a `train_dataset` given to the `Trainer` initialization."):
        trainer.train()


def test_train_amp_save(model: SetFitModel, tmp_path: Path) -> None:
    args = TrainingArguments(output_dir=str(tmp_path), use_amp=True, save_steps=5, num_epochs=5)
    dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2]})
    trainer = Trainer(model, args=args, train_dataset=dataset, eval_dataset=dataset)
    trainer.train()
    assert trainer.evaluate() == {"accuracy": 1.0}
    assert "checkpoint-5" in os.listdir(tmp_path)


def test_evaluate_with_strings(model: SetFitModel) -> None:
    dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": ["positive", "positive", "negative"]})
    trainer = Trainer(model, train_dataset=dataset, eval_dataset=dataset)
    trainer.train()
    metrics = trainer.evaluate()
    assert "accuracy" in metrics
    assert set(model.labels) == {"positive", "negative"}


def test_trainer_wrong_args(model: SetFitModel) -> None:
    dataset = Dataset.from_dict({"text": ["a", "b", "c"], "label": [0, 1, 2]})
    expected = "`args` must be a `TrainingArguments` instance imported from `setfit`."
    with pytest.raises(ValueError, match=expected):
        Trainer(model, dataset)



from unittest import TestCase

import pytest
from transformers import IntervalStrategy

from setfit.training_args import TrainingArguments


class TestTrainingArguments(TestCase):
    def test_raises_error_with_wrong_warmup_proportion(self):
        # warmup_proportion must not be > 1.0
        with pytest.raises(ValueError):
            TrainingArguments(warmup_proportion=1.1)

        # warmup_proportion must not be < 0.0
        with pytest.raises(ValueError):
            TrainingArguments(warmup_proportion=-0.1)

    def test_batch_sizes(self):
        batch_size_A = 12
        batch_size_B = 4

        args = TrainingArguments(batch_size=batch_size_A)
        self.assertEqual(args.batch_size, (batch_size_A, batch_size_A))
        self.assertEqual(args.embedding_batch_size, batch_size_A)
        self.assertEqual(args.classifier_batch_size, batch_size_A)

        args = TrainingArguments(batch_size=(batch_size_A, batch_size_B))
        self.assertEqual(args.batch_size, (batch_size_A, batch_size_B))
        self.assertEqual(args.embedding_batch_size, batch_size_A)
        self.assertEqual(args.classifier_batch_size, batch_size_B)

    def test_num_epochs(self):
        num_epochs_A = 12
        num_epochs_B = 4

        args = TrainingArguments(num_epochs=num_epochs_A)
        self.assertEqual(args.num_epochs, (num_epochs_A, num_epochs_A))
        self.assertEqual(args.embedding_num_epochs, num_epochs_A)
        self.assertEqual(args.classifier_num_epochs, num_epochs_A)

        args = TrainingArguments(num_epochs=(num_epochs_A, num_epochs_B))
        self.assertEqual(args.num_epochs, (num_epochs_A, num_epochs_B))
        self.assertEqual(args.embedding_num_epochs, num_epochs_A)
        self.assertEqual(args.classifier_num_epochs, num_epochs_B)

    def test_learning_rates(self):
        learning_rate_A = 1e-2
        learning_rate_B = 1e-3

        base = TrainingArguments()

        args = TrainingArguments(body_learning_rate=learning_rate_A)
        self.assertEqual(args.body_learning_rate, (learning_rate_A, learning_rate_A))
        self.assertEqual(args.body_embedding_learning_rate, learning_rate_A)
        self.assertEqual(args.body_classifier_learning_rate, learning_rate_A)
        self.assertEqual(args.head_learning_rate, base.head_learning_rate)

        args = TrainingArguments(body_learning_rate=(learning_rate_A, learning_rate_B))
        self.assertEqual(args.body_learning_rate, (learning_rate_A, learning_rate_B))
        self.assertEqual(args.body_embedding_learning_rate, learning_rate_A)
        self.assertEqual(args.body_classifier_learning_rate, learning_rate_B)
        self.assertEqual(args.head_learning_rate, base.head_learning_rate)

    def test_report_to(self):
        args = TrainingArguments(report_to="none")
        self.assertEqual(args.report_to, ["none"])
        args = TrainingArguments(report_to=["none"])
        self.assertEqual(args.report_to, ["none"])
        args = TrainingArguments(report_to="hello")
        self.assertEqual(args.report_to, ["hello"])

    def test_eval_steps_without_eval_strat(self):
        args = TrainingArguments(eval_steps=5)
        self.assertEqual(args.eval_strategy, IntervalStrategy.STEPS)

    def test_eval_strat_steps_without_eval_steps(self):
        args = TrainingArguments(eval_strategy="steps")
        self.assertEqual(args.eval_steps, args.logging_steps)
        with self.assertRaises(ValueError):
            TrainingArguments(eval_strategy="steps", logging_steps=0, logging_strategy="no")

    def test_load_best_model(self):
        with self.assertRaises(ValueError):
            TrainingArguments(load_best_model_at_end=True, eval_strategy="steps", save_strategy="epoch")
        with self.assertRaises(ValueError):
            TrainingArguments(
                load_best_model_at_end=True,
                eval_strategy="steps",
                save_strategy="steps",
                eval_steps=100,
                save_steps=50,
            )
        # No error: save_steps is a round multiple of eval_steps
        TrainingArguments(
            load_best_model_at_end=True,
            eval_strategy="steps",
            save_strategy="steps",
            eval_steps=50,
            save_steps=100,
        )

    def test_logging_steps_zero(self):
        with self.assertRaises(ValueError):
            TrainingArguments(logging_strategy="steps", logging_steps=0)



import json
from pathlib import Path
from unittest import TestCase

import numpy as np
import pytest
import torch
from datasets import Dataset, load_dataset
from sentence_transformers import SentenceTransformer
from sklearn.linear_model import LogisticRegression
from sklearn.multiclass import OneVsRestClassifier
from sklearn.multioutput import ClassifierChain, MultiOutputClassifier

from setfit import SetFitHead, SetFitModel, Trainer
from setfit.modeling import MODEL_HEAD_NAME
from tests.utils import SafeTemporaryDirectory


torch_cuda_available = pytest.mark.skipif(not torch.cuda.is_available(), reason="PyTorch must be compiled with CUDA")


def test_setfit_model_body():
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")

    assert type(model.model_body) is SentenceTransformer


def test_setfit_default_model_head():
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")

    assert type(model.model_head) is LogisticRegression


def test_setfit_model_head_params():
    params = {
        "head_params": {
            "max_iter": 200,
            "solver": "newton-cg",
        }
    }

    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2", **params)

    assert type(model.model_head) is LogisticRegression
    assert params["head_params"] == {
        parameter: value
        for parameter, value in model.model_head.get_params(deep=False).items()
        if parameter in params["head_params"]
    }


def test_setfit_multilabel_one_vs_rest_model_head():
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", multi_target_strategy="one-vs-rest"
    )

    assert type(model.model_head) is OneVsRestClassifier


def test_setfit_multilabel_multi_output_classifier_model_head():
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", multi_target_strategy="multi-output"
    )

    assert type(model.model_head) is MultiOutputClassifier


def test_setfit_multilabel_classifier_chain_classifier_model_head():
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", multi_target_strategy="classifier-chain"
    )

    assert type(model.model_head) is ClassifierChain


class SetFitModelDifferentiableHeadTest(TestCase):
    @classmethod
    def setUpClass(cls):
        dataset = load_dataset("sst2")
        num_classes = 2
        train_dataset = dataset["train"].shuffle(seed=42).select(range(2 * num_classes))
        x_train, y_train = train_dataset["sentence"], train_dataset["label"]

        model = cls._build_model(num_classes)
        model.unfreeze()  # unfreeze the model body and head

        # run one step
        model.model_body.train()
        model.model_head.train()

        dataloader = model._prepare_dataloader(x_train, y_train, batch_size=2 * num_classes)
        criterion = model.model_head.get_loss_fn()
        optimizer = model._prepare_optimizer(2e-4, None, 0.1)
        device = model.model_head.device

        batch = next(iter(dataloader))
        features, labels = batch
        features = {k: v.to(device) for k, v in features.items()}
        labels = labels.to(device)
        optimizer.zero_grad()

        outputs = model.model_body(features)
        outputs = model.model_head(outputs)
        loss = criterion(outputs["logits"], labels)
        loss.backward()
        optimizer.step()

        cls.model = model
        cls.out_features = num_classes
        cls.x_train = x_train
        cls.y_train = y_train

    @staticmethod
    def _build_model(num_classes: int) -> SetFitModel:
        model = SetFitModel.from_pretrained(
            "sentence-transformers/paraphrase-albert-small-v2",
            use_differentiable_head=True,
            head_params={"out_features": num_classes},
        )

        return model

    def test_setfit_body_and_head_on_same_device(self):
        model = self._build_model(num_classes=1)
        assert model.model_body.device.type == model.model_head.device.type

    def test_setfit_single_target_differentiable_head(self):
        model = self._build_model(num_classes=1)

        assert type(model.model_head) is SetFitHead
        assert model.model_head.out_features == 2

    def test_setfit_multi_targets_differentiable_head(self):
        assert type(self.model.model_head) is SetFitHead
        assert self.model.model_head.out_features == self.out_features

    def test_setfit_model_forward(self):
        # Already ran the model's forward in the fixture, so do simple testing here.
        assert type(self.model) is SetFitModel

    def test_setfit_model_backward(self):
        # check the model head's gradients
        for name, param in self.model.model_head.named_parameters():
            assert param.grad is not None, f"Gradients of {name} in the model head is None."
            assert not (param.grad == 0).all().item(), f"All gradients of {name} in the model head are zeros."
            assert not param.grad.isnan().any().item(), f"Gradients of {name} in the model head have NaN."
            assert not param.grad.isinf().any().item(), f"Gradients of {name} in the model head have Inf."

        # check the model body's gradients
        for name, param in self.model.model_body.named_parameters():
            if "0.auto_model.pooler" in name:  # ignore pooler
                continue

            assert param.grad is not None, f"Gradients of {name} in the model body is None."
            assert not (param.grad == 0).all().item(), f"All gradients of {name} in the model body are zeros."
            assert not param.grad.isnan().any().item(), f"Gradients of {name} in the model body have NaN."
            assert not param.grad.isinf().any().item(), f"Gradients of {name} in the model body have Inf."

    def test_max_length_is_larger_than_max_acceptable_length(self):
        max_length = int(1e6)
        dataloader = self.model._prepare_dataloader(self.x_train, self.y_train, batch_size=1, max_length=max_length)

        assert dataloader.dataset.max_length == self.model.model_body.get_max_seq_length()

    def test_max_length_is_smaller_than_max_acceptable_length(self):
        max_length = 32
        dataloader = self.model._prepare_dataloader(self.x_train, self.y_train, batch_size=1, max_length=max_length)

        assert dataloader.dataset.max_length == max_length


def test_setfit_from_pretrained_local_model_without_head(tmp_path):
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    model.save_pretrained(str(tmp_path.absolute()))

    (tmp_path / MODEL_HEAD_NAME).unlink()  # Delete head

    model = SetFitModel.from_pretrained(str(tmp_path.absolute()))

    assert isinstance(model, SetFitModel)


def test_setfit_from_pretrained_local_model_with_head(tmp_path):
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    model.save_pretrained(str(tmp_path.absolute()))

    model = SetFitModel.from_pretrained(str(tmp_path.absolute()))

    assert isinstance(model, SetFitModel)


def test_setfithead_multitarget_from_pretrained():
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2",
        use_differentiable_head=True,
        multi_target_strategy="one-vs-rest",
        head_params={"out_features": 5},
    )
    assert isinstance(model.model_head, SetFitHead)
    assert model.model_head.multitarget
    assert isinstance(model.model_head.get_loss_fn(), torch.nn.BCEWithLogitsLoss)

    y_pred = model.predict("Test text")
    assert len(y_pred) == 5

    y_pred_probs = model.predict_proba("Test text", as_numpy=True)
    assert not np.isclose(y_pred_probs.sum(), 1)  # Should not sum to one


def test_to_logistic_head():
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2")
    devices = (
        [torch.device("cpu"), torch.device("cuda", 0), torch.device("cpu")]
        if torch.cuda.is_available()
        else [torch.device("cpu")]
    )
    for device in devices:
        model.to(device)
        assert model.model_body.device == device


def test_to_torch_head():
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", use_differentiable_head=True
    )
    devices = (
        [torch.device("cpu"), torch.device("cuda", 0), torch.device("cpu")]
        if torch.cuda.is_available()
        else [torch.device("cpu")]
    )
    for device in devices:
        model.to(device)
        assert model.model_body.device == device
        assert model.model_head.device == device


@torch_cuda_available
@pytest.mark.parametrize("use_differentiable_head", [True, False])
def test_to_sentence_transformer_device_reset(use_differentiable_head):
    # This should initialize SentenceTransformer() without a specific device
    # which sets the model to CUDA iff CUDA is available.
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", use_differentiable_head=use_differentiable_head
    )
    # If we move the entire model to CPU, we expect it to stay on CPU forever,
    # Even after encoding or fitting
    model.to("cpu")
    assert model.model_body.device == torch.device("cpu")

    model.model_body.encode("This is a test sample to encode")
    assert model.model_body.device == torch.device("cpu")


@torch_cuda_available
@pytest.mark.parametrize("device", ["cpu", "cuda"])
def test_load_model_on_device(device):
    model = SetFitModel.from_pretrained("sentence-transformers/paraphrase-albert-small-v2", device=device)
    assert model.device.type == device
    assert model.model_body.device.type == device

    model.model_body.encode("This is a test sample to encode")


def test_save_load_config(model: SetFitModel) -> None:
    with SafeTemporaryDirectory() as tmp_dir:
        tmp_dir = str(Path(tmp_dir) / "model")
        model.save_pretrained(tmp_dir)
        config_path = Path(tmp_dir) / "config_setfit.json"
        assert config_path.exists()
        with open(config_path, "r") as f:
            config = json.load(f)
        assert config == {"normalize_embeddings": False, "labels": None}

    with SafeTemporaryDirectory() as tmp_dir:
        tmp_dir = str(Path(tmp_dir) / "model")
        model.normalize_embeddings = True
        model.labels = ["negative", "positive"]
        model.save_pretrained(tmp_dir)
        config_path = Path(tmp_dir) / "config_setfit.json"
        assert config_path.exists()
        with open(config_path, "r") as f:
            config = json.load(f)
        assert config == {"normalize_embeddings": True, "labels": ["negative", "positive"]}

        fresh_model = model.from_pretrained(tmp_dir)
        assert fresh_model.normalize_embeddings is True
        assert fresh_model.labels == ["negative", "positive"]


def test_load_model() -> None:
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2", labels=["foo", "bar", "baz"]
    )
    assert model.labels == ["foo", "bar", "baz"]
    assert model.label2id == {"foo": 0, "bar": 1, "baz": 2}
    assert model.id2label == {0: "foo", 1: "bar", 2: "baz"}


def test_inference_with_labels() -> None:
    model = SetFitModel.from_pretrained("SetFit/test-setfit-sst2")
    assert model.labels is None
    assert model.predict(["Very good"]) == torch.tensor([1], dtype=torch.int32)
    model.labels = ["negative", "positive"]
    assert model.predict(["Very good"]) == ["positive"]

    model = SetFitModel.from_pretrained("SetFit/test-setfit-sst2-string-labels")
    assert model.labels is None
    assert model.predict(["Very good"]) == np.array(["positive"], dtype="<U8")
    model.labels = ["negative", "positive"]
    assert model.predict(["Very good"]) == ["positive"]

    model = SetFitModel.from_pretrained("SetFit/test-setfit-sst2-diff-head")
    assert model.labels is None
    assert model.predict(["Very good"]) == torch.tensor([1], dtype=torch.int32, device=model.device)
    model.labels = ["negative", "positive"]
    assert model.predict(["Very good"]) == ["positive"]


def test_singular_predict() -> None:
    model = SetFitModel.from_pretrained("SetFit/test-setfit-sst2")
    assert model.predict("That was cool!") == torch.tensor(1, dtype=torch.int32)
    probs = model.predict_proba("That was cool!")
    assert probs.shape == (2,)
    assert probs.argmax() == 1
    model.labels = ["negative", "positive"]
    assert model("That was cool!") == "positive"


# A differentiable head may still cause unexpected performance
@pytest.mark.parametrize("use_differentiable_head", [False])
def test_predict_proba_multi_output(use_differentiable_head: bool) -> None:
    model = SetFitModel.from_pretrained(
        "sentence-transformers/paraphrase-albert-small-v2",
        multi_target_strategy="multi-output",
        use_differentiable_head=use_differentiable_head,
    )
    train_dataset = Dataset.from_dict({"text": ["Hello", "World"], "label": [[1, 0], [0, 1]]})

    trainer = Trainer(model=model, train_dataset=train_dataset)
    trainer.train()

    outputs = model.predict_proba("That was cool!")
    assert isinstance(outputs, torch.Tensor)
    assert outputs.shape == (2, 2)
    outputs = model.predict_proba("That was cool!", as_numpy=True)
    assert isinstance(outputs, np.ndarray)
    assert outputs.shape == (2, 2)

    outputs = model.predict_proba(["That was cool!"] * 3)
    assert isinstance(outputs, torch.Tensor)
    assert outputs.shape == (3, 2, 2)
    outputs = model.predict_proba(["That was cool!"] * 3, as_numpy=True)
    assert isinstance(outputs, np.ndarray)
    assert outputs.shape == (3, 2, 2)



import json
import os
import tempfile
import warnings
from pathlib import Path
from typing import Dict, List, Literal, Optional, Set, Tuple, Union

import joblib
import numpy as np
import requests
import torch
from huggingface_hub import ModelHubMixin, hf_hub_download
from huggingface_hub.utils import validate_hf_hub_args
from packaging.version import Version, parse
from sentence_transformers import SentenceTransformer
from sentence_transformers import __version__ as sentence_transformers_version
from sentence_transformers import models
from sklearn.linear_model import LogisticRegression
from sklearn.multiclass import OneVsRestClassifier
from sklearn.multioutput import ClassifierChain, MultiOutputClassifier
from torch import nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm, trange
from transformers.utils import copy_func

from . import logging
from .data import SetFitDataset
from .model_card import SetFitModelCardData, generate_model_card
from .utils import set_docstring


logging.set_verbosity_info()
logger = logging.get_logger(__name__)

MODEL_HEAD_NAME = "model_head.pkl"
CONFIG_NAME = "config_setfit.json"


class SetFitHead(models.Dense):
    """
    A SetFit head that supports multi-class classification for end-to-end training.
    Binary classification is treated as 2-class classification.

    To be compatible with Sentence Transformers, we inherit `Dense` from:
    https://github.com/UKPLab/sentence-transformers/blob/master/sentence_transformers/models/Dense.py

    Args:
        in_features (`int`, *optional*):
            The embedding dimension from the output of the SetFit body. If `None`, defaults to `LazyLinear`.
        out_features (`int`, defaults to `2`):
            The number of targets. If set `out_features` to 1 for binary classification, it will be changed to 2 as 2-class classification.
        temperature (`float`, defaults to `1.0`):
            A logits' scaling factor. Higher values make the model less confident and lower values make
            it more confident.
        eps (`float`, defaults to `1e-5`):
            A value for numerical stability when scaling logits.
        bias (`bool`, *optional*, defaults to `True`):
            Whether to add bias to the head.
        device (`torch.device`, str, *optional*):
            The device the model will be sent to. If `None`, will check whether GPU is available.
        multitarget (`bool`, defaults to `False`):
            Enable multi-target classification by making `out_features` binary predictions instead
            of a single multinomial prediction.
    """

    def __init__(
        self,
        in_features: Optional[int] = None,
        out_features: int = 2,
        temperature: float = 1.0,
        eps: float = 1e-5,
        bias: bool = True,
        device: Optional[Union[torch.device, str]] = None,
        multitarget: bool = False,
    ) -> None:
        super(models.Dense, self).__init__()  # init on models.Dense's parent: nn.Module

        if out_features == 1:
            logger.warning(
                "Change `out_features` from 1 to 2 since we use `CrossEntropyLoss` for binary classification."
            )
            out_features = 2

        if in_features is not None:
            self.linear = nn.Linear(in_features, out_features, bias=bias)
        else:
            self.linear = nn.LazyLinear(out_features, bias=bias)

        self.in_features = in_features
        self.out_features = out_features
        self.temperature = temperature
        self.eps = eps
        self.bias = bias
        self._device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.multitarget = multitarget

        self.to(self._device)
        self.apply(self._init_weight)

    def forward(
        self,
        features: Union[Dict[str, torch.Tensor], torch.Tensor],
        temperature: Optional[float] = None,
    ) -> Union[Dict[str, torch.Tensor], Tuple[torch.Tensor]]:
        """
        SetFitHead can accept embeddings in:
        1. Output format (`dict`) from Sentence-Transformers.
        2. Pure `torch.Tensor`.

        Args:
            features (`Dict[str, torch.Tensor]` or `torch.Tensor):
                The embeddings from the encoder. If using `dict` format,
                make sure to store embeddings under the key: 'sentence_embedding'
                and the outputs will be under the key: 'prediction'.
            temperature (`float`, *optional*):
                A logits' scaling factor. Higher values make the model less
                confident and lower values make it more confident.
                Will override the temperature given during initialization.
        Returns:
            [`Dict[str, torch.Tensor]` or `Tuple[torch.Tensor]`]
        """
        temperature = temperature or self.temperature
        is_features_dict = False  # whether `features` is dict or not
        if isinstance(features, dict):
            assert "sentence_embedding" in features
            is_features_dict = True
        x = features["sentence_embedding"] if is_features_dict else features
        logits = self.linear(x)
        logits = logits / (temperature + self.eps)
        if self.multitarget:  # multiple targets per item
            probs = torch.sigmoid(logits)
        else:  # one target per item
            probs = nn.functional.softmax(logits, dim=-1)
        if is_features_dict:
            features.update(
                {
                    "logits": logits,
                    "probs": probs,
                }
            )
            return features

        return logits, probs

    def predict_proba(self, x_test: torch.Tensor) -> torch.Tensor:
        self.eval()

        return self(x_test)[1]

    def predict(self, x_test: torch.Tensor) -> torch.Tensor:
        probs = self.predict_proba(x_test)

        if self.multitarget:
            return torch.where(probs >= 0.5, 1, 0)
        return torch.argmax(probs, dim=-1)

    def get_loss_fn(self) -> nn.Module:
        if self.multitarget:  # if sigmoid output
            return torch.nn.BCEWithLogitsLoss()
        return torch.nn.CrossEntropyLoss()

    @property
    def device(self) -> torch.device:
        """
        `torch.device`: The device on which the model is placed.

        Reference from: https://github.com/UKPLab/sentence-transformers/blob/master/sentence_transformers/SentenceTransformer.py#L869
        """
        return next(self.parameters()).device

    def get_config_dict(self) -> Dict[str, Optional[Union[int, float, bool]]]:
        return {
            "in_features": self.in_features,
            "out_features": self.out_features,
            "temperature": self.temperature,
            "bias": self.bias,
            "device": self.device.type,  # store the string of the device, instead of `torch.device`
        }

    @staticmethod
    def _init_weight(module) -> None:
        if isinstance(module, nn.Linear):
            nn.init.xavier_uniform_(module.weight)
            if module.bias is not None:
                nn.init.constant_(module.bias, 1e-2)

    def __repr__(self) -> str:
        return "SetFitHead({})".format(self.get_config_dict())


class SetFitModel(ModelHubMixin):
    """A SetFit model with integration to the [Hugging Face Hub](https://huggingface.co).

    Example::

        >>> from setfit import SetFitModel
        >>> model = SetFitModel.from_pretrained("tomaarsen/setfit-bge-small-v1.5-sst2-8-shot")
        >>> model.predict([
        ...     "It's a charming and often affecting journey.",
        ...     "It's slow -- very, very slow.",
        ...     "A sometimes tedious film.",
        ... ])
        ['positive', 'negative', 'negative']
    """

    def __init__(
        self,
        model_body: Optional[SentenceTransformer] = None,
        model_head: Optional[Union[SetFitHead, LogisticRegression]] = None,
        multi_target_strategy: Optional[str] = None,
        normalize_embeddings: bool = False,
        labels: Optional[List[str]] = None,
        model_card_data: Optional[SetFitModelCardData] = None,
        sentence_transformers_kwargs: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        super(SetFitModel, self).__init__()
        self.model_body = model_body
        self.model_head = model_head
        self.multi_target_strategy = multi_target_strategy
        self.normalize_embeddings = normalize_embeddings
        self.labels = labels
        self.model_card_data = model_card_data or SetFitModelCardData()
        self.sentence_transformers_kwargs = sentence_transformers_kwargs or {}

        self.attributes_to_save: Set[str] = {"normalize_embeddings", "labels"}
        self.model_card_data.register_model(self)

    @property
    def has_differentiable_head(self) -> bool:
        # if False, sklearn is assumed to be used instead
        return isinstance(self.model_head, nn.Module)

    @property
    def id2label(self) -> Dict[int, str]:
        """Return a mapping from integer IDs to string labels."""
        if self.labels is None:
            return {}
        return dict(enumerate(self.labels))

    @property
    def label2id(self) -> Dict[str, int]:
        """Return a mapping from string labels to integer IDs."""
        if self.labels is None:
            return {}
        return {label: idx for idx, label in enumerate(self.labels)}

    def fit(
        self,
        x_train: List[str],
        y_train: Union[List[int], List[List[int]]],
        num_epochs: int,
        batch_size: Optional[int] = None,
        body_learning_rate: Optional[float] = None,
        head_learning_rate: Optional[float] = None,
        end_to_end: bool = False,
        l2_weight: Optional[float] = None,
        max_length: Optional[int] = None,
        show_progress_bar: bool = True,
    ) -> None:
        """Train the classifier head, only used if a differentiable PyTorch head is used.

        Args:
            x_train (`List[str]`): A list of training sentences.
            y_train (`Union[List[int], List[List[int]]]`): A list of labels corresponding to the training sentences.
            num_epochs (`int`): The number of epochs to train for.
            batch_size (`int`, *optional*): The batch size to use.
            body_learning_rate (`float`, *optional*): The learning rate for the `SentenceTransformer` body
                in the `AdamW` optimizer. Disregarded if `end_to_end=False`.
            head_learning_rate (`float`, *optional*): The learning rate for the differentiable torch head
                in the `AdamW` optimizer.
            end_to_end (`bool`, defaults to `False`): If True, train the entire model end-to-end.
                Otherwise, freeze the `SentenceTransformer` body and only train the head.
            l2_weight (`float`, *optional*): The l2 weight for both the model body and head
                in the `AdamW` optimizer.
            max_length (`int`, *optional*): The maximum token length a tokenizer can generate. If not provided,
                the maximum length for the `SentenceTransformer` body is used.
            show_progress_bar (`bool`, defaults to `True`): Whether to display a progress bar for the training
                epochs and iterations.
        """
        if self.has_differentiable_head:  # train with pyTorch
            self.model_body.train()
            self.model_head.train()
            if not end_to_end:
                self.freeze("body")

            dataloader = self._prepare_dataloader(list(x_train), list(y_train), batch_size, max_length)
            criterion = self.model_head.get_loss_fn()
            optimizer = self._prepare_optimizer(head_learning_rate, body_learning_rate, l2_weight)
            scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=5, gamma=0.5)
            for epoch_idx in trange(num_epochs, desc="Epoch", disable=not show_progress_bar):
                for batch in tqdm(dataloader, desc="Iteration", disable=not show_progress_bar, leave=False):
                    features, labels = batch
                    optimizer.zero_grad()

                    # to model's device
                    features = {k: v.to(self.device) for k, v in features.items()}
                    labels = labels.to(self.device)

                    outputs = self.model_body(features)
                    if self.normalize_embeddings:
                        outputs["sentence_embedding"] = nn.functional.normalize(
                            outputs["sentence_embedding"], p=2, dim=1
                        )
                    outputs = self.model_head(outputs)
                    logits = outputs["logits"]

                    loss: torch.Tensor = criterion(logits, labels)
                    loss.backward()
                    optimizer.step()

                scheduler.step()

            if not end_to_end:
                self.unfreeze("body")
        else:  # train with sklearn
            embeddings = self.model_body.encode(list(x_train), normalize_embeddings=self.normalize_embeddings)
            self.model_head.fit(embeddings, list(y_train))
            if self.labels is None and self.multi_target_strategy is None:
                # Try to set the labels based on the head classes, if they exist
                # This can fail in various ways, so we catch all exceptions
                try:
                    classes = self.model_head.classes_
                    if classes.dtype.char == "U":
                        self.labels = classes.tolist()
                except Exception:
                    pass

    def _prepare_dataloader(
        self,
        x_train: List[str],
        y_train: Union[List[int], List[List[int]]],
        batch_size: Optional[int] = None,
        max_length: Optional[int] = None,
        shuffle: bool = True,
    ) -> DataLoader:
        max_acceptable_length = self.model_body.get_max_seq_length()
        if max_length is None:
            max_length = max_acceptable_length
            logger.warning(
                f"The `max_length` is `None`. Using the maximum acceptable length according to the current model body: {max_length}."
            )

        if max_length > max_acceptable_length:
            logger.warning(
                (
                    f"The specified `max_length`: {max_length} is greater than the maximum length of the current model body: {max_acceptable_length}. "
                    f"Using {max_acceptable_length} instead."
                )
            )
            max_length = max_acceptable_length

        dataset = SetFitDataset(
            x_train,
            y_train,
            tokenizer=self.model_body.tokenizer,
            max_length=max_length,
        )
        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            collate_fn=dataset.collate_fn,
            shuffle=shuffle,
            pin_memory=True,
        )

        return dataloader

    def _prepare_optimizer(
        self,
        head_learning_rate: float,
        body_learning_rate: Optional[float],
        l2_weight: float,
    ) -> torch.optim.Optimizer:
        body_learning_rate = body_learning_rate or head_learning_rate
        l2_weight = l2_weight or 1e-2
        optimizer = torch.optim.AdamW(
            [
                {
                    "params": self.model_body.parameters(),
                    "lr": body_learning_rate,
                    "weight_decay": l2_weight,
                },
                {"params": self.model_head.parameters(), "lr": head_learning_rate, "weight_decay": l2_weight},
            ],
        )

        return optimizer

    def freeze(self, component: Optional[Literal["body", "head"]] = None) -> None:
        """Freeze the model body and/or the head, preventing further training on that component until unfrozen.

        Args:
            component (`Literal["body", "head"]`, *optional*): Either "body" or "head" to freeze that component.
                If no component is provided, freeze both. Defaults to None.
        """
        if component is None or component == "body":
            self._freeze_or_not(self.model_body, to_freeze=True)

        if (component is None or component == "head") and self.has_differentiable_head:
            self._freeze_or_not(self.model_head, to_freeze=True)

    def unfreeze(
        self, component: Optional[Literal["body", "head"]] = None, keep_body_frozen: Optional[bool] = None
    ) -> None:
        """Unfreeze the model body and/or the head, allowing further training on that component.

        Args:
            component (`Literal["body", "head"]`, *optional*): Either "body" or "head" to unfreeze that component.
                If no component is provided, unfreeze both. Defaults to None.
            keep_body_frozen (`bool`, *optional*): Deprecated argument, use `component` instead.
        """
        if keep_body_frozen is not None:
            warnings.warn(
                "`keep_body_frozen` is deprecated and will be removed in v2.0.0 of SetFit. "
                'Please either pass "head", "body" or no arguments to unfreeze both.',
                DeprecationWarning,
                stacklevel=2,
            )
            # If the body must stay frozen, only unfreeze the head. Eventually, this entire if-branch
            # can be removed.
            if keep_body_frozen and not component:
                component = "head"

        if component is None or component == "body":
            self._freeze_or_not(self.model_body, to_freeze=False)

        if (component is None or component == "head") and self.has_differentiable_head:
            self._freeze_or_not(self.model_head, to_freeze=False)

    def _freeze_or_not(self, model: nn.Module, to_freeze: bool) -> None:
        """Set `requires_grad=not to_freeze` for all parameters in `model`"""
        for param in model.parameters():
            param.requires_grad = not to_freeze

    def encode(
        self, inputs: List[str], batch_size: int = 32, show_progress_bar: Optional[bool] = None
    ) -> Union[torch.Tensor, np.ndarray]:
        """Convert input sentences to embeddings using the `SentenceTransformer` body.

        Args:
            inputs (`List[str]`): The input sentences to embed.
            batch_size (`int`, defaults to `32`): The batch size to use in encoding the sentences to embeddings.
                Higher often means faster processing but higher memory usage.
            show_progress_bar (`Optional[bool]`, defaults to `None`): Whether to show a progress bar while encoding.

        Returns:
            Union[torch.Tensor, np.ndarray]: A matrix with shape [INPUT_LENGTH, EMBEDDING_SIZE], as a
            torch Tensor if this model has a differentiable Torch head, or otherwise as a numpy array.
        """
        return self.model_body.encode(
            inputs,
            batch_size=batch_size,
            normalize_embeddings=self.normalize_embeddings,
            convert_to_tensor=self.has_differentiable_head,
            show_progress_bar=show_progress_bar,
        )

    def _output_type_conversion(
        self, outputs: Union[torch.Tensor, np.ndarray], as_numpy: bool = False
    ) -> Union[torch.Tensor, np.ndarray]:
        """Return `outputs` in the desired type:
        * Numpy array if no differentiable head is used.
        * Torch tensor if a differentiable head is used.

        Note:
            If the model is trained with string labels, which is only possible with a non-differentiable head,
            then we cannot output using torch Tensors, but only using a numpy array.

        Returns:
            Union[torch.Tensor, "ndarray"]: The input, correctly converted to the desired type.
        """
        if as_numpy and self.has_differentiable_head:
            outputs = outputs.detach().cpu().numpy()
        elif not as_numpy and not self.has_differentiable_head and outputs.dtype.char != "U":
            # Only output as tensor if the output isn't a string
            outputs = torch.from_numpy(outputs)
        return outputs

    @torch.no_grad()
    def predict_proba(
        self,
        inputs: Union[str, List[str]],
        batch_size: int = 32,
        as_numpy: bool = False,
        show_progress_bar: Optional[bool] = None,
    ) -> Union[torch.Tensor, np.ndarray]:
        """Predict the probabilities of the various classes.

        Args:
            inputs (`Union[str, List[str]]`): The input sentences to predict class probabilities for.
            batch_size (`int`, defaults to `32`): The batch size to use in encoding the sentences to embeddings.
                Higher often means faster processing but higher memory usage.
            as_numpy (`bool`, defaults to `False`): Whether to output as numpy array instead.
            show_progress_bar (`Optional[bool]`, defaults to `None`): Whether to show a progress bar while encoding.

        Example::

            >>> model = SetFitModel.from_pretrained(...)
            >>> model.predict_proba(["What a boring display", "Exhilarating through and through", "I'm wowed!"])
            tensor([[0.9367, 0.0633],
                    [0.0627, 0.9373],
                    [0.0890, 0.9110]], dtype=torch.float64)
            >>> model.predict_proba("That was cool!")
            tensor([0.8421, 0.1579], dtype=torch.float64)

        Returns:
            `Union[torch.Tensor, np.ndarray]`: A matrix with shape [INPUT_LENGTH, NUM_CLASSES] denoting
            probabilities of predicting an input as a class. If the input is a string, then the output
            is a vector with shape [NUM_CLASSES,].
        """
        is_singular = isinstance(inputs, str)
        if is_singular:
            inputs = [inputs]
        embeddings = self.encode(inputs, batch_size=batch_size, show_progress_bar=show_progress_bar)
        probs = self.model_head.predict_proba(embeddings)
        if isinstance(probs, list):
            if self.has_differentiable_head:
                probs = torch.stack(probs, axis=1)
            else:
                probs = np.stack(probs, axis=1)
        outputs = self._output_type_conversion(probs, as_numpy=as_numpy)
        return outputs[0] if is_singular else outputs

    @torch.no_grad()
    def predict(
        self,
        inputs: Union[str, List[str]],
        batch_size: int = 32,
        as_numpy: bool = False,
        use_labels: bool = True,
        show_progress_bar: Optional[bool] = None,
    ) -> Union[torch.Tensor, np.ndarray, List[str], int, str]:
        """Predict the various classes.

        Args:
            inputs (`Union[str, List[str]]`): The input sentence or sentences to predict classes for.
            batch_size (`int`, defaults to `32`): The batch size to use in encoding the sentences to embeddings.
                Higher often means faster processing but higher memory usage.
            as_numpy (`bool`, defaults to `False`): Whether to output as numpy array instead.
            use_labels (`bool`, defaults to `True`): Whether to try and return elements of `SetFitModel.labels`.
            show_progress_bar (`Optional[bool]`, defaults to `None`): Whether to show a progress bar while encoding.

        Example::

            >>> model = SetFitModel.from_pretrained(...)
            >>> model.predict(["What a boring display", "Exhilarating through and through", "I'm wowed!"])
            ["negative", "positive", "positive"]
            >>> model.predict("That was cool!")
            "positive"

        Returns:
            `Union[torch.Tensor, np.ndarray, List[str], int, str]`: A list of string labels with equal length to the
                inputs if `use_labels` is `True` and `SetFitModel.labels` has been defined. Otherwise a vector with
                equal length to the inputs, denoting to which class each input is predicted to belong. If the inputs
                is a single string, then the output is a single label as well.
        """
        is_singular = isinstance(inputs, str)
        if is_singular:
            inputs = [inputs]
        embeddings = self.encode(list(inputs), batch_size=batch_size, show_progress_bar=show_progress_bar)
        preds = self.model_head.predict(embeddings)
        # If labels are defined, we don't have multilabels & the output is not already strings, then we convert to string labels
        if (
            use_labels
            and self.labels
            and preds.ndim == 1
            and (self.has_differentiable_head or preds.dtype.char != "U")
        ):
            outputs = [self.labels[int(pred)] for pred in preds]
        else:
            outputs = self._output_type_conversion(preds, as_numpy=as_numpy)
        return outputs[0] if is_singular else outputs

    def __call__(
        self,
        inputs: Union[str, List[str]],
        batch_size: int = 32,
        as_numpy: bool = False,
        use_labels: bool = True,
        show_progress_bar: Optional[bool] = None,
    ) -> Union[torch.Tensor, np.ndarray, List[str], int, str]:
        """Predict the various classes.

        Args:
            inputs (`Union[str, List[str]]`): The input sentence or sentences to predict classes for.
            batch_size (`int`, defaults to `32`): The batch size to use in encoding the sentences to embeddings.
                Higher often means faster processing but higher memory usage.
            as_numpy (`bool`, defaults to `False`): Whether to output as numpy array instead.
            use_labels (`bool`, defaults to `True`): Whether to try and return elements of `SetFitModel.labels`.
            show_progress_bar (`Optional[bool]`, defaults to `None`): Whether to show a progress bar while encoding.

        Example::

            >>> model = SetFitModel.from_pretrained(...)
            >>> model(["What a boring display", "Exhilarating through and through", "I'm wowed!"])
            ["negative", "positive", "positive"]
            >>> model("That was cool!")
            "positive"

        Returns:
            `Union[torch.Tensor, np.ndarray, List[str], int, str]`: A list of string labels with equal length to the
                inputs if `use_labels` is `True` and `SetFitModel.labels` has been defined. Otherwise a vector with
                equal length to the inputs, denoting to which class each input is predicted to belong. If the inputs
                is a single string, then the output is a single label as well.
        """
        return self.predict(
            inputs,
            batch_size=batch_size,
            as_numpy=as_numpy,
            use_labels=use_labels,
            show_progress_bar=show_progress_bar,
        )

    @property
    def device(self) -> torch.device:
        """Get the Torch device that this model is on.

        Returns:
            torch.device: The device that the model is on.
        """
        # SentenceTransformers.device is reliable from 2.3.0 onwards
        if parse(sentence_transformers_version) >= Version("2.3.0"):
            return self.model_body.device
        return self.model_body._target_device

    def to(self, device: Union[str, torch.device]) -> "SetFitModel":
        """Move this SetFitModel to `device`, and then return `self`. This method does not copy.

        Args:
            device (Union[str, torch.device]): The identifier of the device to move the model to.

        Example::

            >>> model = SetFitModel.from_pretrained(...)
            >>> model.to("cpu")
            >>> model(["cats are cute", "dogs are loyal"])

        Returns:
            SetFitModel: Returns the original model, but now on the desired device.
        """
        # Note that we must also set _target_device with sentence-transformers <2.3.0,
        # or any SentenceTransformer.fit() call will reset the body location
        if parse(sentence_transformers_version) < Version("2.3.0"):
            self.model_body._target_device = device if isinstance(device, torch.device) else torch.device(device)
        self.model_body = self.model_body.to(device)

        if self.has_differentiable_head:
            self.model_head = self.model_head.to(device)

        return self

    def create_model_card(self, path: str, model_name: Optional[str] = "SetFit Model") -> None:
        """Creates and saves a model card for a SetFit model.

        Args:
            path (str): The path to save the model card to.
            model_name (str, *optional*): The name of the model. Defaults to `SetFit Model`.
        """
        if not os.path.exists(path):
            os.makedirs(path)

        # If the model_path is a folder that exists locally, i.e. when create_model_card is called
        # via push_to_hub, and the path is in a temporary folder, then we only take the last two
        # directories
        model_path = Path(model_name)
        if (
            self.model_card_data.model_id is None
            and model_path.exists()
            and Path(tempfile.gettempdir()) in model_path.resolve().parents
        ):
            self.model_card_data.model_id = "/".join(model_path.parts[-2:])

        with open(os.path.join(path, "README.md"), "w", encoding="utf-8") as f:
            f.write(self.generate_model_card())

    def generate_model_card(self) -> str:
        """Generate and return a model card string based on the model card data.

        Returns:
            str: The model card string.
        """
        return generate_model_card(self)

    def _save_pretrained(self, save_directory: Union[Path, str]) -> None:
        save_directory = str(save_directory)
        # Save the config
        config_path = os.path.join(save_directory, CONFIG_NAME)
        with open(config_path, "w") as f:
            json.dump(
                {
                    attr_name: getattr(self, attr_name)
                    for attr_name in self.attributes_to_save
                    if hasattr(self, attr_name)
                },
                f,
                indent=2,
            )
        # Save the body
        self.model_body.save(path=save_directory, create_model_card=False)
        # Save the README
        self.create_model_card(path=save_directory, model_name=save_directory)
        # Move the head to the CPU before saving
        if self.has_differentiable_head:
            self.model_head.to("cpu")
        # Save the classification head
        joblib.dump(self.model_head, str(Path(save_directory) / MODEL_HEAD_NAME))
        if self.has_differentiable_head:
            self.model_head.to(self.device)

    @classmethod
    @validate_hf_hub_args
    def _from_pretrained(
        cls,
        model_id: str,
        revision: Optional[str] = None,
        cache_dir: Optional[str] = None,
        force_download: Optional[bool] = None,
        proxies: Optional[Dict] = None,
        resume_download: Optional[bool] = None,
        local_files_only: Optional[bool] = None,
        token: Optional[Union[bool, str]] = None,
        multi_target_strategy: Optional[str] = None,
        use_differentiable_head: bool = False,
        device: Optional[Union[torch.device, str]] = None,
        trust_remote_code: bool = False,
        **model_kwargs,
    ) -> "SetFitModel":
        sentence_transformers_kwargs = {
            "cache_folder": cache_dir,
            "use_auth_token": token,
            "device": device,
            "trust_remote_code": trust_remote_code,
        }
        if parse(sentence_transformers_version) >= Version("2.3.0"):
            sentence_transformers_kwargs = {
                "cache_folder": cache_dir,
                "token": token,
                "device": device,
                "trust_remote_code": trust_remote_code,
            }
        else:
            if trust_remote_code:
                raise ValueError(
                    "The `trust_remote_code` argument is only supported for `sentence-transformers` >= 2.3.0."
                )
            sentence_transformers_kwargs = {
                "cache_folder": cache_dir,
                "use_auth_token": token,
                "device": device,
            }
        model_body = SentenceTransformer(model_id, **sentence_transformers_kwargs)
        if parse(sentence_transformers_version) >= Version("2.3.0"):
            device = model_body.device
        else:
            device = model_body._target_device
        model_body.to(device)  # put `model_body` on the target device

        # Try to load a SetFit config file
        config_file: Optional[str] = None
        if os.path.isdir(model_id):
            if CONFIG_NAME in os.listdir(model_id):
                config_file = os.path.join(model_id, CONFIG_NAME)
        else:
            try:
                config_file = hf_hub_download(
                    repo_id=model_id,
                    filename=CONFIG_NAME,
                    revision=revision,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    token=token,
                    local_files_only=local_files_only,
                )
            except requests.exceptions.RequestException:
                pass

        model_kwargs = {key: value for key, value in model_kwargs.items() if value is not None}

        if config_file is not None:
            with open(config_file, "r", encoding="utf-8") as f:
                config = json.load(f)
            # Update model_kwargs + warnings
            for setting, value in config.items():
                if setting in model_kwargs:
                    if model_kwargs[setting] != value:
                        logger.warning(
                            f"Overriding {setting} in model configuration from {value} to {model_kwargs[setting]}."
                        )
                else:
                    model_kwargs[setting] = value

        # Try to load a model head file
        if os.path.isdir(model_id):
            if MODEL_HEAD_NAME in os.listdir(model_id):
                model_head_file = os.path.join(model_id, MODEL_HEAD_NAME)
            else:
                logger.info(
                    f"{MODEL_HEAD_NAME} not found in {Path(model_id).resolve()},"
                    " initialising classification head with random weights."
                    " You should TRAIN this model on a downstream task to use it for predictions and inference."
                )
                model_head_file = None
        else:
            try:
                model_head_file = hf_hub_download(
                    repo_id=model_id,
                    filename=MODEL_HEAD_NAME,
                    revision=revision,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    token=token,
                    local_files_only=local_files_only,
                )
            except requests.exceptions.RequestException:
                logger.info(
                    f"{MODEL_HEAD_NAME} not found on HuggingFace Hub, initialising classification head with random weights."
                    " You should TRAIN this model on a downstream task to use it for predictions and inference."
                )
                model_head_file = None

        model_card_data: SetFitModelCardData = model_kwargs.pop("model_card_data", SetFitModelCardData())

        if model_head_file is not None:
            model_head = joblib.load(model_head_file)
            if isinstance(model_head, torch.nn.Module):
                model_head.to(device)
            model_card_data.infer_st_id(model_id)
        else:
            head_params = model_kwargs.pop("head_params", {})
            if use_differentiable_head:
                if multi_target_strategy is None:
                    use_multitarget = False
                else:
                    if multi_target_strategy in ["one-vs-rest", "multi-output"]:
                        use_multitarget = True
                    else:
                        raise ValueError(
                            f"multi_target_strategy '{multi_target_strategy}' is not supported for differentiable head"
                        )
                # Base `model_head` parameters
                # - get the sentence embedding dimension from the `model_body`
                # - follow the `model_body`, put `model_head` on the target device
                base_head_params = {
                    "in_features": model_body.get_sentence_embedding_dimension(),
                    "device": device,
                    "multitarget": use_multitarget,
                }
                model_head = SetFitHead(**{**head_params, **base_head_params})
            else:
                clf = LogisticRegression(**head_params)
                if multi_target_strategy is not None:
                    if multi_target_strategy == "one-vs-rest":
                        multilabel_classifier = OneVsRestClassifier(clf)
                    elif multi_target_strategy == "multi-output":
                        multilabel_classifier = MultiOutputClassifier(clf)
                    elif multi_target_strategy == "classifier-chain":
                        multilabel_classifier = ClassifierChain(clf)
                    else:
                        raise ValueError(f"multi_target_strategy {multi_target_strategy} is not supported.")

                    model_head = multilabel_classifier
                else:
                    model_head = clf

            model_card_data.set_st_id(model_id if "/" in model_id else f"sentence-transformers/{model_id}")

        # Remove the `transformers` config
        model_kwargs.pop("config", None)
        return cls(
            model_body=model_body,
            model_head=model_head,
            multi_target_strategy=multi_target_strategy,
            model_card_data=model_card_data,
            sentence_transformers_kwargs=sentence_transformers_kwargs,
            **model_kwargs,
        )


docstring = SetFitModel.from_pretrained.__doc__
cut_index = docstring.find("model_kwargs")
if cut_index != -1:
    docstring = (
        docstring[:cut_index]
        + """labels (`List[str]`, *optional*):
                If the labels are integers ranging from `0` to `num_classes-1`, then these labels indicate
                    the corresponding labels.
            model_card_data (`SetFitModelCardData`, *optional*):
                A `SetFitModelCardData` instance storing data such as model language, license, dataset name,
                    etc. to be used in the automatically generated model cards.
            multi_target_strategy (`str`, *optional*):
                The strategy to use with multi-label classification. One of "one-vs-rest", "multi-output",
                    or "classifier-chain".
            use_differentiable_head (`bool`, *optional*):
                Whether to load SetFit using a differentiable (i.e., Torch) head instead of Logistic Regression.
            normalize_embeddings (`bool`, *optional*):
                Whether to apply normalization on the embeddings produced by the Sentence Transformer body.
            device (`Union[torch.device, str]`, *optional*):
                The device on which to load the SetFit model, e.g. `"cuda:0"`, `"mps"` or `torch.device("cuda")`.
            trust_remote_code (`bool`, defaults to `False`): Whether or not to allow for custom Sentence Transformers
                models defined on the Hub in their own modeling files. This option should only be set to True for
                repositories you trust and in which you have read the code, as it will execute code present on
                the Hub on your local machine. Defaults to False.

        Example::

            >>> from setfit import SetFitModel
            >>> model = SetFitModel.from_pretrained(
            ...     "sentence-transformers/paraphrase-mpnet-base-v2",
            ...     labels=["positive", "negative"],
            ... )
        """
    )
    SetFitModel.from_pretrained = set_docstring(SetFitModel.from_pretrained, docstring)

SetFitModel.save_pretrained = copy_func(SetFitModel.save_pretrained)
SetFitModel.save_pretrained.__doc__ = SetFitModel.save_pretrained.__doc__.replace(
    "~ModelHubMixin._from_pretrained", "SetFitModel.push_to_hub"
)
